import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv


myCop1 = ot.GumbelCopula(2)
myCop2 = ot.NormalCopula(2)
alpha = 0.3
myOrdSumCop = ot.OrdinalSumCopula([myCop1, myCop2], [alpha])
myOrdSumCop.setDescription(["$u_1$", "$u_2$"])
graphPDF = myOrdSumCop.drawPDF()
graphCDF = myOrdSumCop.drawCDF()


fig = plt.figure(figsize=(8, 4))
pdf_axis = fig.add_subplot(121)
cdf_axis = fig.add_subplot(122)
pdf_axis.set_xlim(auto=True)
cdf_axis.set_xlim(auto=True)

otv.View(graphPDF, figure=fig, axes=[pdf_axis], add_legend=True)
otv.View(graphCDF, figure=fig, axes=[cdf_axis], add_legend=True)
fig.suptitle("Ordinal Sum of Copulas: Gumbel(2) and Normal(2): pdf and cdf")



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

distribution = ot.Rayleigh(2.5, -1.0)
size = 10000
sample = distribution.getSample(size)
factory = ot.RayleighFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", estimatedDistribution)
estimatedRayleigh = factory.buildAsRayleigh(sample)
print("Rayleigh          =", distribution)
print("Estimated rayleigh=", estimatedRayleigh)
estimatedRayleigh = factory.buildAsRayleigh()
print("Default rayleigh=", estimatedRayleigh)
estimatedRayleigh = factory.buildAsRayleigh(distribution.getParameter())
print("Rayleigh from parameters=", estimatedRayleigh)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

dimension = 2
epsilon = ot.Point(dimension, 1e-4)
x = ot.Point(dimension, 2.0)

step = ot.ConstantStep(epsilon)
print("step type=", step.getClassName(), "step value=", step(x))

eta = ot.Point(dimension, 1.0)
step = ot.BlendedStep(epsilon, eta)
print("step type=", step.getClassName(), "step value=", step(x))



"""
Model a singular multivariate distribution
==========================================
"""

# %%
# The objective of the example is to show how to fit a singular copula
# on a sample of dimension :math:`n_\inputDim` which might be high. For example,
# this is the case when we want to fit the joint distribution of the :math:`n_\inputDim` Karhunen Loeve
# coefficients in the Karhunen Loeve modelisation of a process.
#
# We first show that when the copula is singular, using a kernel smoothing fitting might
# not be a good idea.
#
# The good way to follow is to fit an empirical Bernstein copula implemented in
# :class:`~openturns.EmpiricalBernsteinCopula` through its factory
# :class:`~openturns.BernsteinCopulaFactory`. This factory garantees that the built
# :class:`~openturns.EmpiricalBernsteinCopula` is really a copula and not only a
# core (ie a multivariate distribution which range is included in :math:`[0,1]^\inputDim`).
# According to the level of singularity, the number :math:`m` of cells used to build the
# empirical Bernstein copula differs: the higher the singularity, the larger :math:`m`.
#
# The cells partition the sample: all the points of the same cell are grouped into a
# single *global* point. The empirical Bernstein copula is a mixture of products of Beta
# distributions centered on the global point of each cell.
#
# We denote by :math:`\sampleSize` the sample size. The number of global points varies according to the number of
# cells considered (math:`m`):
#
# - :math:`m = \sampleSize` means that all the sample has been retained: we create one cell around
#   each point of the sample; in that case, the empirical Bernstein copula is the *Beta copula* in the sens of [segers2016]_;
# - :math:`m = 1` means the sample has been grouped into one single global point: we get the independent copula;
# - :math:`1 < m < \sampleSize` means that we create :math:`m` cells gathering several points.
#
# As the  empirical Bernstein copula defined in this way is a copula only if  :math:`m`  divides
# :math:`\sampleSize`, the :class:`~openturns.BernsteinCopulaFactory` throws away
# some part of the sample is set aside in order to check this condition.
#
# A point's *zone of influence* is its cell. The larger :math:`m`, the smaller its zone of influence:
# the final copula thus adheres strongly to the sample. The smaller :math:`m`, the larger
# its zone of influence, and the more the final copula will be able to fill a large
# area around it.
#
# So, if we know that the final copula is diffuse, we recommend to let the
# :class:`~openturns.BernsteinCopulaFactory` automatically calculate the optimal number of cells to take. If
# we know that the final copula is singular, we recommend to specify the value of :math:`m` to be equal
# to the sample size :math:`\sampleSize`.
#
# We illustrate the influence of  :math:`m` on a singular copula.

import openturns as ot
import openturns.viewer as otv
import math as m

# %%
# Routine to draw a distribution cloud and a sample.


def draw(dist, Y):
    g = ot.Graph()
    g.setAxes(True)
    g.setGrid(True)
    c = ot.Cloud(dist.getSample(10000))
    c.setColor("red")
    c.setPointStyle("bullet")
    g.add(c)
    c = ot.Cloud(Y)
    c.setColor("black")
    c.setPointStyle("bullet")
    g.add(c)
    g.setBoundingBox(
        ot.Interval(
            Y.getMin() - 0.5 * Y.computeRange(), Y.getMax() + 0.5 * Y.computeRange()
        )
    )
    return g


# %%
# Introduction
# ~~~~~~~~~~~~
#
# We consider the function :math:`f: \Rset^3 \rightarrow \Rset` defined by:
#
# .. math::
#
#      f(u, v_1, v_2) = (y_1, y_2)
#
# where:
#
# .. math::
#
#      y_1 & = \sin(u) / (1 + \cos(u)^2) + 0.05 * v_1 \\
#      y_2 & = \sin(u) \cos(u) / (1 + \cos(u)^2) + 0.05 * v_2
#
# We define the following input random vector:
#
# .. math::
#
#      U  \sim \cU(-0.85\pi, 0.85\pi) \\
#      (V_1, V_2)  \sim \cN(\vect{\mu} = \vect{0}, \vect{\sigma} = \vect{1}, Id_2)\\
#
# with :math:`U` and :math:`\vect{V})` independent.
#
# We define the output random vector :math:`\vect{Y}` as:
#
# .. math::
#
#      \vect{Y} = f(U, V_1, V_2)

f = ot.SymbolicFunction(
    ["U", "v1", "v2"],
    ["sin(U)/(1+cos(U)^2)+0.05*v1", "sin(U)*cos(U)/(1+cos(U)^2)+0.05*v2"],
)
U = ot.Uniform(-0.85 * m.pi, 0.85 * m.pi)
V = ot.Normal(2)
X = ot.BlockIndependentDistribution([U, V])

# %%
# We generate a sample of the output random vector :math:`\vect{Y}` of size :math:`N`.
N = 200
sample_Y = f(X.getSample(N))

# %%
# Multivariate kernel smoothing
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# We estimate the distribution of the output random vector :math:`\vect{Y}` by multivariate kernel smoothing.
y_multi_ks = ot.KernelSmoothing().build(sample_Y)
view = otv.View(draw(y_multi_ks, sample_Y))

# %%
# We see that the fitting is not satisfactory: the empty regions have not been respected. The fitted copula is
# diffuse and does not model the observed singularity correctly.

# %%
# Empirical Bernstein copula factory
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
#
# Now, we estimate the distribution of :math:`\vect{Y}` splitting the estimation of the marginals
# from the estimation of the copula:
#
# - the marginals are fitted by kernel smoothing,
# - the copula is fitted using the :class:`~openturns.BernsteinCopulaFactory` that builds
#   an empirical Bernstein copula.
#
# First, we do not specify the bin number :math:`m`. It is equal to the value computed by the default method, which is the
# LogLikelihood criteria.
empBern_copula = ot.BernsteinCopulaFactory().buildAsEmpiricalBernsteinCopula(sample_Y)
print("bin number computed m = ", empBern_copula.getBinNumber())
marginals = [
    ot.KernelSmoothing().build(sample_Y.getMarginal(j))
    for j in range(sample_Y.getDimension())
]
y_empBern = ot.JointDistribution(marginals, empBern_copula)
view = otv.View(draw(y_empBern, sample_Y))

# %%
# We see that the optimal number of cells is :math:`m = 1`: it means that one single cell has been created.
# The built copula  is the independent copula. It is diffuse
# in :math:`[0,1]^2`. This is not satisfactory. The optimal
# :math:`m` is not correct fot the singular copula we try to estimate.

# %%
# Now, we specify a bin number equal to the sample size: :math:`m = N` so that the built copula is very close to the sample.
empBern_copula = ot.BernsteinCopulaFactory().build(sample_Y, N)
y_empBern = ot.JointDistribution(marginals, empBern_copula)
view = otv.View(draw(y_empBern, sample_Y))
otv.View.ShowAll()

# %%

# In that case, the built copula manages to reproduce the specific feature we observe in the data.



"""
Save/load a study
=================
The objective of this example is to demonstrate how to save the structures created within a script session to disk in order to be able to load them in a future session.

There are several possible ways to achieve this:

- with the standard pickle module
- with openturns's Study

"""

# %%
import openturns as ot
import pickle


# %%
# create objects to save
distribution = ot.Normal(4.0, 3.0)
function = ot.SymbolicFunction(["x1", "x2"], ["x1 + x2"])

# %%
# **With the pickle module**
#
# The objects are retrieved in the same order they are stored.

# %%
# save objects
with open("study.pkl", "wb") as f:
    pickle.dump(distribution, f)
    pickle.dump(function, f)

# %%
# load saved objects
with open("study.pkl", "rb") as f:
    loaded_distribution = pickle.load(f)
    loaded_function = pickle.load(f)
str(loaded_distribution), str(loaded_function)

# %%
# **With OpenTURNS' Study**
#
# In order to be able to manipulate the objects contained in a Study, it is necessary to:
#
# - create the same empty structure in the new study,
# - fill this new empty structure with the content of the loaded structure, identified with its name or its id.
#
# Each object is identified whether with:
#
# - its name: it is useful to give names to the objects we want to save. If no name has been given by the user, we can use the default name.
#   The name of each saved object can be checked in the output XML file or with the python `print` command (applied to the `Study` object).
# - its id number: this id number is unique to each object. It distinguishes objects with identical type and name (like the default name "Unnamed").
#   This id number may be checked by printing the study **after** it has been loaded in the python interface (with the `print` command).
#   It can differ from the id number indicated in the XML file the study was loaded from.
# - for HDF5 storage (see below): the id serves both as xml id and hdf5 dataset name.
#   Id uniqueness forbids any misleading in reading/writing hdf5 datasets.

# %%
# Create a Study Object
study = ot.Study()

# %%
# Associate it to an XML file
fileName = "study.xml"
study.setStorageManager(ot.XMLStorageManager(fileName))

# %%
# Alternatively, large amounts of data can be stored in binary HDF5 file.
# An XML file (`study_h5.xml`) serves as header for binary data, which are stored in the automatically created `study_h5.h5` file.
study_h5 = ot.Study()
fileName_h5 = "study_h5.xml"
study_h5.setStorageManager(ot.XMLH5StorageManager(fileName_h5))

# %%
# Add an object to the study; at this point it is not written to disk yet
study.add("distribution", distribution)
study.add("function", function)

# %%
# Save the study; this writes into the file
study.save()

# %%
# Create a new study associated to the same file
study = ot.Study()
study.setStorageManager(ot.XMLStorageManager(fileName))

# %%
# Load the file and all its objects
study.load()

# %%
# Check the content of the myStudy
print("Study = ", study)

# %%
# List names of stored objects
study.getLabels()

# %%
# Check our 'distribution' labelled object was loaded
study.hasObject("distribution")

# %%
# Load the objects; we must create a void object of the desired type (or parent type)
distribution2 = ot.Normal()
function2 = ot.Function()
study.fillObject("distribution", distribution2)
study.fillObject("function", function2)
str(distribution2), str(function2)



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott


parameters = ot.GammaMuSigma(0.1, 0.489898, -0.5)
distribution = ot.ParametrizedDistribution(parameters)

print("Distribution ", distribution)

# Is this distribution elliptical ?
print("Elliptical = ", distribution.isElliptical())

# Is this distribution continuous ?
print("Continuous = ", distribution.isContinuous())

# Test for realization of distribution
oneRealization = distribution.getRealization()
print("oneRealization=", oneRealization)

# Define a point
point = ot.Point(distribution.getDimension(), 1.0)
print("Point= ", point)

# Show PDF and CDF of point
DDF = distribution.computeDDF(point)
print("ddf     =", DDF)
LPDF = distribution.computeLogPDF(point)
print("log pdf= %.6g" % LPDF)
PDF = distribution.computePDF(point)
print("pdf     =%.6g" % PDF)
CDF = distribution.computeCDF(point)
print("cdf= %.6g" % CDF)
CCDF = distribution.computeComplementaryCDF(point)
print("ccdf= %.6g" % CCDF)
Survival = distribution.computeSurvivalFunction(point)
print("survival= %.6g" % Survival)
CF = distribution.computeCharacteristicFunction(point[0])
print("characteristic function=(%.6g, %.6g)" % (CF.real, CF.imag))
LCF = distribution.computeLogCharacteristicFunction(point[0])
print("log characteristic function=(%.6g, %.6g)" % (LCF.real, LCF.imag))
PDFgr = distribution.computePDFGradient(point)
quantile = distribution.computeQuantile(0.95)
print("quantile=", quantile)
print("cdf(quantile)=", distribution.computeCDF(quantile))
# Get 95% survival function
inverseSurvival = ot.Point(distribution.computeInverseSurvivalFunction(0.95))
print("InverseSurvival=", repr(inverseSurvival))
print(
    "Survival(inverseSurvival)=%.6f"
    % distribution.computeSurvivalFunction(inverseSurvival)
)

# Confidence regions
interval, threshold = distribution.computeMinimumVolumeIntervalWithMarginalProbability(
    0.95
)
ott.assert_almost_equal(interval.getLowerBound(), [-0.49937], 1e-4, 0.0)
ott.assert_almost_equal(interval.getUpperBound(), [1.06337], 1e-4, 0.0)
print("threshold=", ot.Point(1, threshold))
levelSet, beta = distribution.computeMinimumVolumeLevelSetWithThreshold(0.95)
print("Minimum volume level set=", levelSet)
print("beta=", ot.Point(1, beta))
interval, beta = distribution.computeBilateralConfidenceIntervalWithMarginalProbability(
    0.95
)
print("Bilateral confidence interval=", interval)
print("beta=", ot.Point(1, beta))
(
    interval,
    beta,
) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(0.95, False)
print("Unilateral confidence interval (lower tail)=", interval)
print("beta=", ot.Point(1, beta))
(
    interval,
    beta,
) = distribution.computeUnilateralConfidenceIntervalWithMarginalProbability(0.95, True)
print("Unilateral confidence interval (upper tail)=", interval)
print("beta=", ot.Point(1, beta))

mean = distribution.getMean()
print("mean=", mean)
covariance = distribution.getCovariance()
print("covariance=", covariance)
correlation = distribution.getCorrelation()
print("correlation=", correlation)
spearman = distribution.getSpearmanCorrelation()
print("spearman=", spearman)
kendall = distribution.getKendallTau()
print("kendall=", kendall)
parameters = distribution.getParametersCollection()
print("parameters=", parameters)
print("Standard representative=", distribution.getStandardRepresentative())

cpdf = distribution.computeConditionalPDF(0.0, [])
print(f"conditional PDF={cpdf:.6f}")
scpdf = distribution.computeSequentialConditionalPDF([0.0])
print(f"sequential conditional PDF={scpdf}")
ccdf = distribution.computeConditionalCDF(0.0, [])
print(f"conditional CDF={ccdf:.6f}")
sccdf = distribution.computeSequentialConditionalCDF([0.0])
print(f"sequential conditional CDF={sccdf}")

dp = distribution.getDistributionParameters()

ot.RandomGenerator.SetSeed(2)
ot.Log.Show(ot.Log.TRACE)
validation = ott.DistributionValidation(distribution)
validation.run()



#! /usr/bin/env python

import openturns as ot
from openturns.testing import assert_almost_equal
from openturns.usecases import stiffened_panel as stiffened_panel


ot.TESTPREAMBLE()
ot.PlatformInfo.SetNumericalPrecision(15)

print("Test the StiffenedPanel class.")
sp = stiffened_panel.StiffenedPanel()

print(sp.distribution.getMean())
# test parameter
assert sp.dim == 10

# test marginals means
assert_almost_equal(sp.youngModulus.getMean()[0], 1.1e11, 1e-12)
assert_almost_equal(sp.nu.getMean()[0], 0.375, 1e-12)
assert_almost_equal(sp.h_c.getMean()[0], 0.03, 1e-12)
assert_almost_equal(sp.ell.getMean()[0], 0.049, 1e-12)
assert_almost_equal(sp.f_1.getMean()[0], 0.028, 1e-12)
assert_almost_equal(sp.f_2.getMean()[0], 0.0066, 1e-12)
assert_almost_equal(sp.t.getMean()[0], 8.1005e-5, 1e-12)
assert_almost_equal(sp.a.getMean()[0], 0.61, 1e-12)
assert_almost_equal(sp.b_0.getMean()[0], 0.045, 1e-12)
assert_almost_equal(sp.p.getMean()[0], 0.038, 1e-12)

meanInput = sp.distribution.getMean()
print("meanInput = ", meanInput)
meanOutput = sp.model(meanInput)
print("meanOutput = ", meanOutput)
assert_almost_equal(meanOutput, [197.76173070456073], 1.0e-15)



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

mesh = ot.IntervalMesher([9]).build(ot.Interval(-1.0, 1.0))
cov1D = ot.AbsoluteExponential([1.0])
algo = ot.KarhunenLoeveP1Algorithm(mesh, cov1D, 0.0)
algo.setNbModes(8)
algo.run()
result = algo.getResult()
lambd = result.getEigenvalues()
KLModes = result.getModesAsProcessSample()
print("KL modes=", KLModes)
print("KL eigenvalues=", lambd)
process = ot.GaussianProcess(cov1D, KLModes.getMesh())
coefficients = result.project(process.getSample(10))
print("KL coefficients=", coefficients)
KLFunctions = result.getModes()
print("KL functions=", KLFunctions)
print("KL lift=", result.lift(coefficients[0]))
print("KL lift as field=", result.liftAsField(coefficients[0]))

# check spectra variant
if ot.PlatformInfo.HasFeature("spectra"):
    ot.ResourceMap.SetAsString("KarhunenLoeveP1Algorithm-EigenvaluesSolver", "SPECTRA")
    algo.run()
    result2 = algo.getResult()
    lambd2 = result2.getEigenvalues()
    ott.assert_almost_equal(lambd2, lambd)
    ot.ResourceMap.SetAsString("KarhunenLoeveP1Algorithm-EigenvaluesSolver", "LAPACK")

R = ot.CorrelationMatrix(2)
R[0, 1] = 0.5
scale = [1.0]
amplitude = [1.0, 2.0]
cov2D = ot.ExponentialModel(scale, amplitude, R)
algo = ot.KarhunenLoeveP1Algorithm(mesh, cov2D, 0.0)
algo.setNbModes(18)
algo.run()
result = algo.getResult()
lambd = result.getEigenvalues()
KLModes = result.getModesAsProcessSample()
print("KL modes=", KLModes)
print("KL eigenvalues=", lambd)
process = ot.GaussianProcess(cov2D, KLModes.getMesh())
coefficients = result.project(process.getSample(10))
print("KL coefficients=", coefficients)
KLFunctions = result.getModes()
print("KL functions=", KLFunctions)
print("KL lift=", result.lift(coefficients[0]))
print("KL lift as field=", result.liftAsField(coefficients[0]))

# 2d input, to check that issue #1660 is solved
mesh = ot.IntervalMesher([4] * 2).build(ot.Interval([-1.2] * 2, [1.0] * 2))
cov2D = ot.AbsoluteExponential([1.0] * 2)
algo = ot.KarhunenLoeveP1Algorithm(mesh, cov2D, 1e-3)
algo.run()
result = algo.getResult()
lambd = result.getEigenvalues()
KLModesPS = result.getModesAsProcessSample()
# The output is hidden due to near-zero nonreproducible values
# print("KL modes (process sample)=", KLModesPS)
KLScaledModesPS = result.getScaledModesAsProcessSample()
# The output is hidden due to near-zero nonreproducible values
# print("KL scaled modes (process sample)=", KLScaledModesPS)
KLModes = result.getModes()
# The output is hidden due to near-zero nonreproducible values
# print("KL modes (functions)=", KLModes)
KLScaledModes = result.getScaledModes()
# The output is hidden due to near-zero nonreproducible values
# print("KL scaled modes (functions)=", KLScaledModes)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


# We create a Function
myFunction = ot.SymbolicFunction(
    ["x1", "x2", "x3", "x4"],
    ["(x1*x1+x2^3*x1)/(2*x3*x3+x4^4+1)", "cos(x2*x2+x4)/(x1*x1+1+x3^4)"],
)

# We create a distribution
dim = myFunction.getInputDimension()
R = ot.CorrelationMatrix(dim)
for i in range(dim):
    R[i, i] = 1.0
for i in range(1, dim):
    R[i, i - 1] = 0.5

m = ot.Point(dim, 1.0)
s = ot.Point(dim, 2.0)
distribution = ot.Normal(m, s, R)
ref_distribution = distribution
print("distribution = ", repr(ref_distribution))

# We create a distribution-based RandomVector
X = ot.RandomVector(distribution)
print("X=", X)
print("is composite? ", X.isComposite())

# Check standard methods of class RandomVector
print("X dimension=", X.getDimension())
print("X realization (first )=", repr(X.getRealization()))
print("X realization (second)=", repr(X.getRealization()))
print("X realization (third )=", repr(X.getRealization()))
print("X sample =", repr(X.getSample(5)))

# We create a composite RandomVector Y from X and myFunction
Y = ot.CompositeRandomVector(myFunction, X)
print("Y=", Y)
print("is composite? ", Y.isComposite())

# Check standard methods of class RandomVector
print("Y dimension=", Y.getDimension())
print("Y realization (first )=", repr(Y.getRealization()))
print("Y realization (second)=", repr(Y.getRealization()))
print("Y realization (third )=", repr(Y.getRealization()))
print("Y sample =", repr(Y.getSample(5)))
parameter = Y.getParameter()
print("Y parameter =", repr(parameter))
Y.setParameter(parameter)
print("Y parameter desc =", repr(Y.getParameterDescription()))

# check composite of composite
composite2 = ot.CompositeRandomVector(ot.SymbolicFunction(["y0", "y1"], ["y0+y1"]), Y)
print(composite2)
assert composite2.getAntecedent().getDistribution() == distribution



"""
Create a threshold event
========================
"""

# %%
# Abstract
# --------
#
# We present in this example the creation and the use of a :class:`~openturns.ThresholdEvent` to estimate a simple integral.
#
import openturns as ot
import openturns.viewer as otv

# %%
# We consider a standard Gaussian random vector :math:`X` and build a random vector from this distribution.
distX = ot.Normal()
vecX = ot.RandomVector(distX)

# %%
# We consider the simple model :math:`f:x \mapsto |x|` and consider the output random variable :math:`Y = f(X)`.
f = ot.SymbolicFunction(["x1"], ["abs(x1)"])
vecY = ot.CompositeRandomVector(f, vecX)

# %%
# We define a very simple :class:`~openturns.ThresholdEvent` which happpens whenever :math:`|X| < 1.0` :
thresholdEvent = ot.ThresholdEvent(vecY, ot.Less(), 1.0)

# %%
# For the normal distribution, it is a well-known fact that the values lower than one standard deviation (here exactly 1)
# away from the mean (here 0) account roughly for 68.27% of the set.
# So the probability of the event is:
#
print("Probability of the event : %.4f" % 0.6827)

# %%
# We can also use a basic estimator to get the probability of the event by drawing samples from the initial distribution `distX` and counting those which realize the event:
print(
    "Probability of the event (event sampling) : %.4f"
    % thresholdEvent.getSample(1000).computeMean()[0]
)


# %%
# The geometric interpretation is simply the area under the PDF of the standard normal distribution for :math:`x \in [-1,1]` which we draw thereafter.


# %%
def linearSample(xmin, xmax, npoints):
    """
    Returns a sample created from a regular grid
    from xmin to xmax with npoints points.
    """
    step = (xmax - xmin) / (npoints - 1)
    rg = ot.RegularGrid(xmin, step, npoints)
    vertices = rg.getVertices()
    return vertices


# %%
# The boundary of the event are the lines :math:`x = -1.0` and :math:`x = 1.0`
a, b = -1, 1

# %%
nplot = 100  # Number of points in the plot
x = linearSample(a, b, nplot)
y = distX.computePDF(x)


vLow = [0.0 for i in range(nplot)]
vUp = [y[i, 0] for i in range(nplot)]
area = distX.computeCDF(b) - distX.computeCDF(a)
boundsPoly = ot.Polygon.FillBetween(x.asPoint(), vLow, vUp)

# %%
# We add the colored area to the PDF graph.
graph = distX.drawPDF()
graph.add(boundsPoly)
graph.setTitle("Probability of the event E = %.4f" % (area))
graph.setLegends([""])
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



"""
Fit a parametric copula
=======================
"""

# %%
# In this example we are going to estimate the parameters of a Gaussian copula from a sample.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# Create data
R = ot.CorrelationMatrix(2)
R[1, 0] = 0.4
copula = ot.NormalCopula(R)
sample = copula.getSample(500)

# %%
# Estimate a Normal copula
distribution = ot.NormalCopulaFactory().build(sample)
print(distribution)

# %%
# The estimated parameters
distribution.getParameter()

# %%
# Draw fitted distribution
graph = distribution.drawPDF()
view = otv.View(graph)

# %%
# Display all figures
otv.View.ShowAll()



"""
Calibration of the flooding model
=================================

In this example we are interested in the calibration of the :ref:`flooding model <use-case-flood-model>`.
We calibrate the parameters of a flooding model where only the difference between the
downstream and upstream riverbed levels can be calibrated.
This example shows how to manage the lack of identifiability in a calibration problem.

This example use least squares to calibrate the parametric
model.
Please read :ref:`code_calibration` for more details on code calibration and least squares.
This study is relatively complicated: please read the :doc:`calibration of the Chaboche mechanical model
</auto_calibration/least_squares_and_gaussian_calibration/plot_calibration_chaboche>` first if this is not already done.
The observations that we use in this study are simulated with the script
:doc:`Generate flooding model observations
</auto_calibration/least_squares_and_gaussian_calibration/plot_generate_flooding>`.
"""

# %%
# Parameters to calibrate and observations
# ----------------------------------------
#
# The variables of the model are:
#
# - :math:`Q` : Input. Observed.
# - :math:`K_s`, :math:`Z_v`, :math:`Z_m` : Input. Calibrated.
# - :math:`H`: Output. Observed.
#
# The vector of parameters to calibrate is:
#
# .. math::
#    \theta = (K_s,Z_v,Z_m).
#
# In the description of the :ref:`flooding model<use-case-flood-model>`,
# we see that only one parameter can be identified.
# Hence, calibrating this model requires some regularization.
# We return to this topic when analyzing the singular values of
# the Jacobian matrix.
#
# We consider a sample size equal to:
#
# .. math::
#    n = 10.
#
#
# The observations are the couples :math:`\{(Q_i,H_i)\}_{i=1,...,n}`, i.e. each observation is a
# couple made of the flowrate and the corresponding river height.

from openturns.usecases import flood_model
from matplotlib import pyplot as plt
import openturns.viewer as otv
import numpy as np
import openturns as ot

ot.ResourceMap.SetAsUnsignedInteger("Normal-SmallDimension", 1)

# %%
# Define the observations
# -----------------------
# In practice, we generally use a data set which has been obtained from
# measurements.
# This data set can be loaded using e.g. :meth:`~openturns.Sample.ImportFromCSVFile`.
# Here we import the data from the
# :class:`~openturns.usecases.flood_model.FloodModel`
# class.
fm = flood_model.FloodModel()
Qobs = fm.data[:, 0]
Hobs = fm.data[:, 1]
nbobs = fm.data.getSize()
fm.data

# %%
# Create the physical model
# -------------------------
# We define the model :math:`g` which has 4 inputs and one output H.
#
# The nonlinear least squares algorithm does not take into account for bounds
# in the parameters.
# Therefore, we ensure that the output is computed whatever the inputs.
# The model fails into two situations:
#
# * if :math:`K_s<0`,
# * if :math:`Z_v-Z_m<0`.
#
# In these cases, we return an infinite number, so that the optimization
# algorithm does not get trapped.

# %%


def functionFlooding(X):
    L = 5.0e3
    B = 300.0
    Q, K_s, Z_v, Z_m = X
    alpha = (Z_m - Z_v) / L
    if alpha < 0.0 or K_s <= 0.0:
        H = np.inf
    else:
        H = (Q / (K_s * B * np.sqrt(alpha))) ** (3.0 / 5.0)
    return [H]


g = ot.PythonFunction(4, 1, functionFlooding)
g = ot.MemoizeFunction(g)
g.setInputDescription(["Q ($m^3/s$)", "Ks ($m^{1/3}/s$)", "Zv (m)", "Zm (m)"])
g.setOutputDescription(["H (m)"])

# %%
# Setting the calibration parameters
# ----------------------------------
# Define the value of the reference values of the :math:`\theta` parameter.
# In the Bayesian framework, this is called the mean of the *prior* normal
# distribution.
# In the data assimilation framework, this is called the *background*.
KsInitial = 20.0
ZvInitial = 49.0
ZmInitial = 51.0
thetaPrior = [KsInitial, ZvInitial, ZmInitial]

# %%
# Create the parametric function
# ------------------------------
# In the physical model, the inputs and parameters are ordered as
# presented in the next table.
# Notice that there are no parameters in the physical model.
#
# +-------+----------------+
# | Index | Input variable |
# +=======+================+
# | 0     | Q              |
# +-------+----------------+
# | 1     | Ks             |
# +-------+----------------+
# | 2     | Zv             |
# +-------+----------------+
# | 3     | Zm             |
# +-------+----------------+
#
# +-------+-----------+
# | Index | Parameter |
# +=======+===========+
# | ∅     | ∅         |
# +-------+-----------+
#
# **Table 1.** Indices and names of the inputs and parameters of the physical model.
#
print("Physical Model Inputs:", g.getInputDescription())
print("Physical Model Parameters:", g.getParameterDescription())

# %%
# In order to perform calibration, we have to define a parametric model,
# with observed inputs and parameters to calibrate.
# In order to do this, we create a :class:`~openturns.ParametricFunction`
# where the parameters are `Ks`, `Zv` and `Zm` which have the indices 1, 2
# and 3 in the physical model.
#
# +-------+----------------+
# | Index | Input variable |
# +=======+================+
# | 0     | Q              |
# +-------+----------------+
#
# +-------+-----------+
# | Index | Parameter |
# +=======+===========+
# | 0     | Ks        |
# +-------+-----------+
# | 1     | Zv        |
# +-------+-----------+
# | 2     | Zm        |
# +-------+-----------+
#
# **Table 2.** Indices and names of the inputs and parameters of the parametric model.
#


# %%
# The following statement creates the calibrated function from the model.
# The calibrated parameters :math:`K_s`, :math:`Z_v`, :math:`Z_m` are at
# indices 1, 2, 3 in the inputs arguments of the model.
calibratedIndices = [1, 2, 3]
mycf = ot.ParametricFunction(g, calibratedIndices, thetaPrior)

# %%
# Plot the Y observations versus the X observations.
graph = ot.Graph("Observations", "Q ($m^3/s$)", "H (m)")
# Plot the model before calibration
curve = mycf.draw(100.0, 4000.0).getDrawable(0)
curve.setLegend("Model, before calibration")
curve.setLineStyle(ot.ResourceMap.GetAsString("CalibrationResult-ObservationLineStyle"))
graph.add(curve)
# Plot the noisy observations
cloud = ot.Cloud(Qobs, Hobs)
cloud.setLegend("Observations")
cloud.setPointStyle(
    ot.ResourceMap.GetAsString("CalibrationResult-ObservationPointStyle")
)
graph.add(cloud)
#
graph.setLegendPosition("upper left")
view = otv.View(graph)

# %%
# We see that the model does not fit to the data.
# The goal of calibration is to find which parameter best fit to the
# observations.

# %%
# Calibration with linear least squares
# -------------------------------------
# The :class:`~openturns.LinearLeastSquaresCalibration` class performs the linear
# least squares calibration by linearizing the model in the neighbourhood of
# the reference point.
algo = ot.LinearLeastSquaresCalibration(mycf, Qobs, Hobs, thetaPrior, "SVD")

# %%
# The :meth:`~openturns.LinearLeastSquaresCalibration.run` method computes the
# solution of the problem.
algo.run()
calibrationResult = algo.getResult()

# %%
# The :meth:`~openturns.CalibrationResult.getParameterMAP` method returns the
# maximum of the posterior distribution of :math:`\theta`.
thetaMAP = calibrationResult.getParameterMAP()
print("theta After = ", thetaMAP)
print("theta Before = ", thetaPrior)
# %%
# Print the true values of the parameters.
print("True theta")
print("  Ks = ", fm.trueKs)
print("  Zv = ", fm.trueZv)
print("  Zm = ", fm.trueZm)

# %%
# In this case, we see that there seems to be a great distance from the
# reference value of :math:`\theta` to the optimum: the values seem too large in magnitude.
# As we are going to see, there is an identification problem because the
# Jacobian matrix is rank-degenerate.

# %%
# Diagnostic of the identification issue
# --------------------------------------

# %%
# In this section, we show how to diagnose the identification problem.
#
# The :meth:`~openturns.CalibrationResult.getParameterPosterior` method returns
# the posterior normal distribution of :math:`\theta`.

# %%
distributionPosterior = calibrationResult.getParameterPosterior()
print(distributionPosterior)

# %%
# We see that there is a large covariance matrix diagonal.
#
# Let us compute a 95% confidence interval for the solution :math:`\theta^\star`.

# %%
print(
    distributionPosterior.computeBilateralConfidenceIntervalWithMarginalProbability(
        0.95
    )[0]
)

# %%
# The confidence interval is *very* large.
# In order to clarify the situation, we compute the Jacobian matrix of the
# model at the candidate point.

# %%
mycf.setParameter(thetaPrior)
thetaDim = len(thetaPrior)
jacobianMatrix = ot.Matrix(nbobs, thetaDim)
for i in range(nbobs):
    jacobianMatrix[i, :] = mycf.parameterGradient(Qobs[i]).transpose()
print(jacobianMatrix[0:5, :])

# %%
# The rank of the problem can be seen from the singular values of the Jacobian
# matrix.

# %%
print(jacobianMatrix.computeSingularValues())

# %%
# We can see that there are two singular values which are relatively close to zero.
#
# This explains why the Jacobian matrix is close to being rank-degenerate.
#
# Moreover, this allows one to compute the actual dimensionality of the problem.
# The algorithm we use computes the singular values in descending order.
# Moreover, by definition, the singular values are nonnegative.
# We see that the first singular value is close to :math:`10`
# and the others are very close to :math:`0` in comparison.
# This implies that the (numerical) rank of the Jacobian matrix is 1,
# even if there are 3 parameters.
#
# Hence, only one parameter can be identified, be it :math:`K_s`, :math:`Z_v`
# or :math:`Z_m`.
# The choice of the particular parameter to identify is free.
# However, in hydraulic studies, the parameter :math:`K_s` is classically
# calibrated while :math:`Z_v` and :math:`Z_m` are left constant.

# %%
# Conclusion of the linear least squares calibration
# --------------------------------------------------

# %%
# There are several methods to solve the problem.
#
# * Given that the problem is not identifiable, we can use some regularization
#   method. Two methods are provided in the library: the Gaussian linear least
#   squares `GaussianLinearCalibration` and the Gaussian non linear least
#   squares `GaussianNonlinearCalibration`.
# * We can change the problem, replacing it with a problem which is identifiable.
#   In the flooding model, we can view :math:`Z_v` and :math:`Z_m` as
#   constants and calibrate :math:`K_s` only.
#
# In this example, we do not change the problem and see how the different
# methods perform.

# %%
# Calibration with non linear least squares
# -----------------------------------------
# The :class:`~openturns.NonLinearLeastSquaresCalibration` class performs the
# non linear least squares calibration by minimizing the squared Euclidian norm
# between the predictions and the observations.

# %%
algo = ot.NonLinearLeastSquaresCalibration(mycf, Qobs, Hobs, thetaPrior)

# %%
# The :meth:`~openturns.NonLinearLeastSquaresCalibration.run` method computes
# the solution of the problem.
algo.run()
calibrationResult = algo.getResult()

# %%
# Analysis of the results
# -----------------------

# %%
# The :meth:`~openturns.CalibrationResult.getParameterMAP` method returns the
# maximum of the posterior distribution of :math:`\theta`.

# %%
thetaMAP = calibrationResult.getParameterMAP()
print(thetaMAP)

# %%
# We can compute a 95% confidence interval of the parameter :math:`\theta^\star`.
#
# This confidence interval is based on bootstrap, based on a sample size equal
# to 100 (as long as the value of the :class:`~openturns.ResourceMap` key
# "NonLinearLeastSquaresCalibration-BootstrapSize" is unchanged).
# This confidence interval reflects the sensitivity of the optimum
# to the variability in the observations.
print(
    ot.ResourceMap.GetAsUnsignedInteger(
        "NonLinearLeastSquaresCalibration-BootstrapSize"
    )
)

# %%
thetaPosterior = calibrationResult.getParameterPosterior()
print(thetaPosterior.computeBilateralConfidenceIntervalWithMarginalProbability(0.95)[0])

# %%
# In this case, the values of the parameters are quite accurately
# computed.

# %%
# Increase the default number of points in the plots.
# This produces smoother spiky distributions.
ot.ResourceMap.SetAsUnsignedInteger("Distribution-DefaultPointNumber", 300)

# %%
graph = calibrationResult.drawObservationsVsInputs()
graph.setLegendPosition("upper left")
view = otv.View(graph)

# %%
# We see that there is a good fit after calibration, since the predictions after calibration
# are close to the observations.

# %%
graph = calibrationResult.drawObservationsVsPredictions()
view = otv.View(graph)

# %%
# We see that there is a much better fit after calibration, since the
# predictions are close to the diagonal of the graphics.

# %%
observationError = calibrationResult.getObservationsError()
print(observationError)

# %%
# We can see that the observation error has a sample mean close to zero and a
# sample standard deviation approximately equal to 0.11.

# sphinx_gallery_thumbnail_number = 5
graph = calibrationResult.drawResiduals()
graph.setLegendPosition("upper left")
view = otv.View(graph)

# %%
# The analysis of the residuals shows that the distribution is centered on
# zero and symmetric.
# This indicates that the calibration performed well.
# Moreover, the distribution of the residuals is close to being Gaussian.
# This is an important hypothesis of the least squares method so that
# checking that this hypothesis occurs in the study is an important
# verification.

# %%
graph = calibrationResult.drawParameterDistributions()
view = otv.View(
    graph,
    figure_kw={"figsize": (8.0, 3.0)},
    legend_kw={"bbox_to_anchor": (1.0, 1.0), "loc": "upper left"},
)
plt.subplots_adjust(right=0.8, bottom=0.2)


# %%
def plotDistributionGridPDF(distribution):
    """
    Plot the marginal and bi-dimensional iso-PDF on a grid.

    Parameters
    ----------
    distribution : ot.Distribution
        The distribution.

    Returns
    -------
    grid : ot.GridLayout(dimension, dimension)
        The grid of plots.

    """
    dimension = distribution.getDimension()
    grid = ot.GridLayout(dimension, dimension)
    for i in range(dimension):
        for j in range(dimension):
            if i == j:
                distributionI = distribution.getMarginal([i])
                graph = distributionI.drawPDF()
            else:
                distributionJI = distribution.getMarginal([j, i])
                graph = distributionJI.drawPDF()
            graph.setLegends([""])
            graph.setTitle("")
            if i < dimension - 1:
                graph.setXTitle("")
            if j > 0:
                graph.setYTitle("")
            grid.setGraph(i, j, graph)
    grid.setTitle("Iso-PDF values")
    return grid


# %%
# Plot the PDF values of the distribution of the optimum parameters.
grid = plotDistributionGridPDF(thetaPosterior)
view = otv.View(
    grid,
    figure_kw={"figsize": (9.0, 6.0)},
    legend_kw={"bbox_to_anchor": (1.0, 1.0), "loc": "upper left"},
)
plt.subplots_adjust(wspace=1.0, hspace=0.5)

# %%
# Gaussian linear calibration
# ---------------------------

# %%
# The standard deviation of the observations.
sigmaH = 0.5  # (m^2)

# %%
# Define the covariance matrix of the output Y of the model.
errorCovariance = ot.CovarianceMatrix(1)
errorCovariance[0, 0] = sigmaH**2

# %%
# Define the covariance matrix of the parameters :math:`\theta` to calibrate.
sigmaKs = 5.0
sigmaZv = 1.0
sigmaZm = 1.0
#
sigma = ot.CovarianceMatrix(3)
sigma[0, 0] = sigmaKs**2
sigma[1, 1] = sigmaZv**2
sigma[2, 2] = sigmaZm**2
print(sigma)

# %%
# The :class:`~openturns.GaussianLinearCalibration` class performs Gaussian
# linear calibration by linearizing the model in the neighbourhood of the prior.
algo = ot.GaussianLinearCalibration(
    mycf, Qobs, Hobs, thetaPrior, sigma, errorCovariance, "SVD"
)

# %%
# The :meth:`~openturns.GaussianLinearCalibration.run` method computes
# the solution of the problem.
algo.run()
calibrationResult = algo.getResult()

# %%
# Analysis of the results
# -----------------------
# The :meth:`~openturns.CalibrationResult.getParameterMAP` method returns the
# maximum of the posterior distribution of :math:`\theta`.
thetaMAP = calibrationResult.getParameterMAP()
print(thetaMAP)

# %%
graph = calibrationResult.drawObservationsVsInputs()
graph.setLegendPosition("upper left")
view = otv.View(graph)

# %%
# We see that the output of the model after calibration is closer to the
# observations.

# %%
graph = calibrationResult.drawObservationsVsPredictions()
view = otv.View(graph)

# %%
# In this case, the fit is satisfactory after calibration.

# %%
graph = calibrationResult.drawResiduals()
graph.setLegendPosition("upper left")
view = otv.View(graph)

# %%
# We see that the distribution of the residual is centered on zero after
# calibration.

# %%
# The :meth:`~openturns.CalibrationResult.getParameterPosterior` method
# returns the posterior normal distribution of :math:`\theta`.

# %%
distributionPosterior = calibrationResult.getParameterPosterior()
print(distributionPosterior)

# %%
# We can compute a 95% credibility interval of the parameter :math:`\theta^\star`.
print(
    distributionPosterior.computeBilateralConfidenceIntervalWithMarginalProbability(
        0.95
    )[0]
)

# %%
# We see that there is a large uncertainty on the value of the parameter
# :math:`K_s` which can be as small as :math:`14` and as large as :math:`34`.

# %%
# We can compare the prior and posterior distributions of the marginals of :math:`\theta`.

# %%
graph = calibrationResult.drawParameterDistributions()
view = otv.View(
    graph,
    figure_kw={"figsize": (8.0, 3.0)},
    legend_kw={"bbox_to_anchor": (1.0, 1.0), "loc": "upper left"},
)
plt.subplots_adjust(right=0.8, bottom=0.2)

# %%
# The two distributions are different, which shows that the calibration is
# sensitive to the observations (if the observations were not sensitive, the
# two distributions were superimposed).
# Moreover, the two distributions are quite close, which implies that the prior
# distribution has played a role in the calibration (otherwise the two
# distributions would be completely different,
# indicating that only the observations were taken into account).

# %%
# Plot the PDF values of the distribution of the optimum parameters.
grid = plotDistributionGridPDF(thetaPosterior)
view = otv.View(
    grid,
    figure_kw={"figsize": (9.0, 6.0)},
    legend_kw={"bbox_to_anchor": (1.0, 1.0), "loc": "upper left"},
)
plt.subplots_adjust(wspace=1.0, hspace=0.5)

# %%
# Gaussian nonlinear calibration
# ------------------------------
# The :class:`~openturns.GaussianNonLinearCalibration` class performs Gaussian
# nonlinear calibration.
algo = ot.GaussianNonLinearCalibration(
    mycf, Qobs, Hobs, thetaPrior, sigma, errorCovariance
)

# %%
# The :meth:`~openturns.GaussianNonLinearCalibration.run` method computes the
# solution of the problem.
algo.run()
calibrationResult = algo.getResult()

# %%
# Analysis of the results
# -----------------------
# The :meth:`~openturns.CalibrationResult.getParameterMAP` method returns the
# maximum of the posterior distribution of :math:`\theta`.
thetaMAP = calibrationResult.getParameterMAP()
print(thetaMAP)

# %%
graph = calibrationResult.drawObservationsVsInputs()
graph.setLegendPosition("upper left")
view = otv.View(graph)

# %%
# We see that the output of the model after calibration is in the middle of the
# observations: the calibration seems correct.

# %%
graph = calibrationResult.drawObservationsVsPredictions()
view = otv.View(graph)

# %%
# The fit is excellent after calibration. Indeed, the cloud of points after calibration is on the diagonal.

# %%
graph = calibrationResult.drawResiduals()
view = otv.View(
    graph,
    figure_kw={"figsize": (8.0, 4.0)},
    legend_kw={"bbox_to_anchor": (1.0, 1.0), "loc": "upper left"},
)
plt.subplots_adjust(right=0.6)

# %%
# We see that the distribution of the residual is centered on zero.
# This is a proof that the calibration did perform correctly.

# %%
# The :meth:`~openturns.CalibrationResult.getParameterPosterior` method
# returns the posterior normal distribution of :math:`\theta`.
distributionPosterior = calibrationResult.getParameterPosterior()

# %%
# We can compute a 95% credibility interval of the parameter :math:`\theta^\star`.
print(
    distributionPosterior.computeBilateralConfidenceIntervalWithMarginalProbability(
        0.95
    )[0]
)

# %%
# We see that there is a small uncertainty on the value of all parameters.

# %%
# We can compare the prior and posterior distributions of the marginals of :math:`\theta`.
graph = calibrationResult.drawParameterDistributions()
view = otv.View(
    graph,
    figure_kw={"figsize": (8.0, 3.0)},
    legend_kw={"bbox_to_anchor": (1.0, 1.0), "loc": "upper left"},
)
plt.subplots_adjust(right=0.8, bottom=0.2)

# %%
# The two distributions are very different, with a spiky posterior distribution.
# This shows that the calibration is very sensitive to the observations.

# %%
# Plot the PDF values of the distribution of the optimum parameters.
grid = plotDistributionGridPDF(thetaPosterior)
view = otv.View(
    grid,
    figure_kw={"figsize": (9.0, 6.0)},
    legend_kw={"bbox_to_anchor": (1.0, 1.0), "loc": "upper left"},
)
plt.subplots_adjust(wspace=1.0, hspace=0.5)

# %%
# Tuning the posterior distribution estimation
# --------------------------------------------
#
# The "GaussianNonLinearCalibration-BootstrapSize" key of the
# :class:`~openturns.ResourceMap` controls the posterior distribution estimation.
#
# * If "GaussianNonLinearCalibration-BootstrapSize" > 0 (by default it is equal to 100),
#   then a bootstrap resample algorithm is used to see the dispersion of the MAP estimator.
#   This allows one to see the variability of the estimator with respect to
#   the finite noisy observation sample.
# * If "GaussianNonLinearCalibration-BootstrapSize" is zero, then the
#   Gaussian linear calibration estimator is used (i.e. the :class:`~openturns.GaussianLinearCalibration`
#   class) at the optimum. This is called the Laplace approximation.
#

# %%
# The default value of the key is nonzero, meaning that bootstrap is used.
# This can be costly in some cases, because it requires to repeat the
# optimization several times.
print(ot.ResourceMap.GetAsUnsignedInteger("GaussianNonLinearCalibration-BootstrapSize"))

# %%
# We must configure the key before creating the object (otherwise changing
# the parameter does not change the result).
ot.ResourceMap.SetAsUnsignedInteger("GaussianNonLinearCalibration-BootstrapSize", 0)
algo = ot.GaussianNonLinearCalibration(
    mycf, Qobs, Hobs, thetaPrior, sigma, errorCovariance
)
algo.run()
calibrationResult = algo.getResult()
graph = calibrationResult.drawParameterDistributions()
view = otv.View(
    graph,
    figure_kw={"figsize": (8.0, 3.0)},
    legend_kw={"bbox_to_anchor": (1.0, 1.0), "loc": "upper left"},
)
plt.subplots_adjust(right=0.8, bottom=0.2)

# %%
# As we can see, this does not change much the posterior distribution, which remains spiky.

# %%
# Plot the PDF values of the distribution of the optimum parameters.
grid = plotDistributionGridPDF(thetaPosterior)
view = otv.View(
    grid,
    figure_kw={"figsize": (9.0, 6.0)},
    legend_kw={"bbox_to_anchor": (1.0, 1.0), "loc": "upper left"},
)
plt.subplots_adjust(wspace=1.0, hspace=0.5)

# %%
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot
from openturns.testing import assert_almost_equal
from openturns.usecases import ishigami_function

ot.TESTPREAMBLE()

im = ishigami_function.IshigamiModel()
expectedCoefficientsLinear = [
    3.5,
    1.62542,
    0,
    0,
    0,
    0,
    0,
    -0.594723,
    0,
    0,
    -1.29064,
    0,
    0,
    0,
    0,
    1.37242,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    -1.95229,
    0,
    0,
    0,
    0,
    0.194929,
    0,
    0,
    0,
    0,
    -1.08975,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0.409177,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    1.35741,
    0,
    0,
    0,
    0,
    0,
    0,
    -0.0126684,
    0,
    0,
    0,
    0,
    0.164588,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    -0.324901,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    -0.33939,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0.00046142,
    0,
    0,
    0,
    0,
    -0.0106965,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0.0490707,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0.0459147,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
]
expectedCoefficientsCondensed = [
    3.5,
    1.62542,
    -0.594723,
    -1.29064,
    1.37242,
    -1.95229,
    0.194929,
    -1.08975,
    0.409177,
    1.35741,
    -0.0126684,
    0.164588,
    -0.324901,
    -0.33939,
    0.00046142,
    -0.0106965,
    0.0490707,
    0.0459147,
]
condensedIndices = [
    0,
    1,
    7,
    10,
    15,
    30,
    35,
    40,
    49,
    77,
    84,
    89,
    98,
    156,
    165,
    170,
    179,
    275,
]
expectedCoefficientsHyper = [
    3.5,
    1.62542,
    0,
    0,
    0,
    -0.594723,
    0,
    -1.29064,
    0,
    0,
    0,
    0,
    0,
    0,
    -1.95229,
    0,
    0.194929,
    0,
    0,
    0,
    0,
    0,
    0,
    1.37242,
    0,
    0,
    1.35741,
    0,
    -0.0126684,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    -0.33939,
    0,
    0,
    0,
    0,
    0,
    0,
    0.409177,
    0,
    0.00046142,
    0,
    0,
    0,
    -1.08975,
    0,
    0,
    0,
    0,
    0,
    0.0459147,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
]

# Extract the relevant elements of the Ishigami test-case
im = ishigami_function.IshigamiModel()
dimension = im.dim
model = im.model

marginals = [im.X1, im.X2, im.X3]
distribution = im.distribution

# Create the orthogonal basis
polynomialCollection = [ot.LegendreFactory()] * dimension

enumerateFunction = ot.LinearEnumerateFunction(dimension)
productBasis = ot.OrthogonalProductPolynomialFactory(
    polynomialCollection, enumerateFunction
)

# Create the doe
degree = 10
basisSize = enumerateFunction.getBasisSizeFromTotalDegree(degree)
samplingSize = 32768
marginalSize = degree + 5

doeList = [
    ot.LowDiscrepancyExperiment(
        ot.LowDiscrepancySequence(ot.SobolSequence()), distribution, samplingSize
    ),
    ot.GaussProductExperiment(distribution, [marginalSize] * dimension),
]

for doe in doeList:
    # Sampling
    inputSample, weights = doe.generateWithWeights()
    wMin = min(weights)
    wMax = max(weights)
    outputSample = model(inputSample)
    # Create the polynomial chaos algorithm using the full constructor
    algo = ot.IntegrationExpansion(
        inputSample, weights, outputSample, distribution, productBasis, basisSize
    )
    algo.run()
    # Check the coefficients
    result = algo.getResult()
    assert not result.isLeastSquares()
    assert not result.involvesModelSelection()
    coeffs = result.getCoefficients().asPoint()
    ref = expectedCoefficientsLinear[: coeffs.getSize()]
    err = (coeffs - ref).norm()
    rtol = 5.0e-2
    atol = 5.0e-2
    assert_almost_equal(err, 0.0, rtol, atol)
    # Check the function restriction
    algo.setActiveFunctions(condensedIndices)
    print("algo=", algo)
    algo.run()
    result = algo.getResult()
    coeffs = result.getCoefficients().asPoint()
    ref = expectedCoefficientsCondensed[: coeffs.getSize()]
    err = (coeffs - ref).norm()
    assert_almost_equal(err, 0.0, rtol, atol)
    if wMin == wMax:
        # Create the polynomial chaos algorithm using the full constructor
        algo = ot.IntegrationExpansion(
            inputSample, outputSample, distribution, productBasis, basisSize
        )
        algo.run()
        # Check the coefficients
        result = algo.getResult()
        coeffs = result.getCoefficients().asPoint()
        ref = expectedCoefficientsLinear[: coeffs.getSize()]
        err = (coeffs - ref).norm()
        assert_almost_equal(err, 0.0, rtol, atol)
    # Create the polynomial chaos algorithm using the simplified constructor
    algo = ot.IntegrationExpansion(inputSample, weights, outputSample, distribution)
    algo.run()
    # Check the coefficients
    result = algo.getResult()
    coeffs = result.getCoefficients().asPoint()
    ref = expectedCoefficientsHyper[: coeffs.getSize()]
    err = (coeffs - ref).norm()
    rtol = 1.0e-2
    atol = 1.0e-2
    assert_almost_equal(err, 0.0, rtol, atol)
    # Check the constructors assuming uniform weights
    if wMin == wMax:
        # Create the polynomial chaos algorithm using the simplified constructor
        algo = ot.IntegrationExpansion(inputSample, outputSample, distribution)
        algo.run()
        # Check the coefficients
        result = algo.getResult()
        coeffs = result.getCoefficients().asPoint()
        ref = expectedCoefficientsHyper[: coeffs.getSize()]
        err = (coeffs - ref).norm()
        rtol = 1.0e-2
        atol = 1.0e-2
        assert_almost_equal(err, 0.0, rtol, atol)



"""
Estimate Sobol' indices for the Ishigami function by a sampling method: a quick start guide to sensitivity analysis
===================================================================================================================
"""

# %%
#
# In this example, we estimate the Sobol' indices for the
# :ref:`Ishigami function <use-case-ishigami>` by sampling methods.
#

# %%
# Introduction
# ------------
#
# In this example we are going to quantify the correlation between the input
# variables and the output variable of a model thanks to Sobol indices.
#
# Sobol indices are designed to evaluate the importance of a single variable
# or a specific set of variables.
# Here the Sobol indices are estimated by sampling from the distributions of
# the input variables and propagating uncertainty through a function.
#
# In theory, Sobol indices range from 0 to 1; the closer an index value is
# to 1, the better the associated input variable explains the function output.
#
# Let us denote by :math:`d` the input dimension of the model.
#
# Sobol' indices can be computed at different orders.
#
# * First order indices evaluate the importance of one input variable
#   at a time.
#
# * Total indices give the relative importance of one input variable
#   and all its interactions with other variables.
#   Alternatively, they can be viewed as measuring how much wriggle room
#   remains to the output when all but one input variables are fixed.
#
# * In general, we are only interested in first order and total Sobol' indices.
#   There are situations, however, where we want to estimate interactions.
#   Second order indices evaluate the importance of every pair of input variables.
#   The number of second order indices is:
#
# .. math::
#    \binom{d}{2} = \frac{d \times \left( d-1\right)}{2}.
#
# In practice, when the number of input variables is not small (say,
# when :math:`d>5`), then the number of second order indices is too large
# to be easily analyzed.
# In these situations, we limit the analysis to the first order and total
# Sobol' indices.

# %%
# Define the model
# ----------------

# %%
from openturns.usecases import ishigami_function
import openturns as ot
import openturns.viewer
import openturns.viewer as otv

# %%
# We load the Ishigami model from the usecases model :
im = ishigami_function.IshigamiModel()

# %%
# The :class:`~openturns.usecases.ishigami_function.IshigamiModel` data class contains the input distribution
# :math:`\vect{X}=(X_1, X_2, X_3)` in `im.distribution` and the Ishigami
# function in `im.model`.
# We also have access to the input variable names with:
input_names = im.distribution.getDescription()

# %%
# Draw the function
# -----------------

# %%
n = 10000
sampleX = im.distribution.getSample(n)
sampleY = im.model(sampleX)

# %%
# Display relationships between the output and the inputs

# %%
grid = ot.VisualTest.DrawPairsXY(sampleX, sampleY)
_ = otv.View(grid, figure_kw={"figsize": (10.0, 4.0)})

# %%
graph = ot.HistogramFactory().build(sampleY).drawPDF()
view = otv.View(graph)

# %%
# We see that the distribution of the output has two modes.

# %%
# Estimate the Sobol' indices
# ---------------------------

# %%
# We first create a design of experiments with the `SobolIndicesExperiment`.
# Since we are not interested in second order indices for the moment,
# we use the default value of the third argument (we will come back to this
# topic later).

# %%
size = 1000
sie = ot.SobolIndicesExperiment(im.distribution, size)
inputDesign = sie.generate()
input_names = im.distribution.getDescription()
inputDesign.setDescription(input_names)
inputDesign.getSize()

# %%
# We see that 5000 function evaluations are required to estimate the first
# order and total Sobol' indices.

# %%
# Then we evaluate the outputs corresponding to this design of experiments.

# %%
outputDesign = im.model(inputDesign)

# %%
# Then we estimate the Sobol' indices with the `SaltelliSensitivityAlgorithm`.

# %%
sensitivityAnalysis = ot.SaltelliSensitivityAlgorithm(inputDesign, outputDesign, size)

# %%
# Let us estimate first order and total Sobol' indices.

# %%
sensitivityAnalysis.getFirstOrderIndices()

# %%
sensitivityAnalysis.getTotalOrderIndices()

# %%
# In the following graph, the vertical bars represent
# the 95% confidence intervals of the estimates.

# %%
graph = sensitivityAnalysis.draw()
view = otv.View(graph)

# %%
# - We see that the variable :math:`X_1`, with a total Sobol' index close
#   to 0.6, is the most significant variable, taking into account both its direct
#   effect and its interactions with other variables.
#   Its first order index is close to 0.3, which implies that its interactions
#   alone produce almost 30% (0.6 - 0.3) of the total variance.
# - The variable :math:`X_2` has the highest first order index: approximately 0.4.
#   However, it has little interaction with other variables since its total
#   order index is close to its first order index.
# - The variable :math:`X_3` has a first order index close to zero.
#   However, it has an impact on the total variance thanks to its interactions
#   with :math:`X_1`.
# - We see that the variability of the estimates is quite high even with a
#   relatively large sample size.
#   Moreover, since the exact first order Sobol' index for :math:`X_3` is zero,
#   its estimate has a 50% chance of being nonpositive.

# %%
# Estimate the second order indices
# ---------------------------------

# %%
size = 1000
computeSecondOrder = True
sie = ot.SobolIndicesExperiment(im.distribution, size, computeSecondOrder)
inputDesign = sie.generate()
print(inputDesign.getSize())
inputDesign.setDescription(input_names)
outputDesign = im.model(inputDesign)

# %%
# We see that 8000 function evaluations are now required; that is 3000 more
# evaluations than in the previous situation.

# %%
sensitivityAnalysis = ot.SaltelliSensitivityAlgorithm(inputDesign, outputDesign, size)

# %%
second_order = sensitivityAnalysis.getSecondOrderIndices()
for i in range(im.dim):
    for j in range(i):
        print("2nd order index (%d,%d)=%g" % (i, j, second_order[i, j]))

# %%
# This shows that the only significant interaction is the one between :math:`X_1`
# and :math:`X_3` (beware of Python's index shift: 0 denotes the first input variable).

# %%
# Using a different estimator
# ---------------------------
#
# We have used the `SaltelliSensitivityAlgorithm` class to estimate the indices.
# Others are available in the library:
#
# * `SaltelliSensitivityAlgorithm`
# * `MartinezSensitivityAlgorithm`
# * `JansenSensitivityAlgorithm`
# * `MauntzKucherenkoSensitivityAlgorithm`
#

# %%
# In order to compare the results with another method, we use the
# `MartinezSensitivityAlgorithm` class.

# %%
sensitivityAnalysis = ot.MartinezSensitivityAlgorithm(inputDesign, outputDesign, size)

# %%
graph = sensitivityAnalysis.draw()
view = otv.View(graph)

# %%
# We see that the results do not change significantly in this particular situation.

# %%
# Display all figures
otv.View.ShowAll()



import sympy as sp

theta_, u, v = sp.symbols("theta_ u v")


thetaMinus1_ = theta_ - 1
uv = thetaMinus1_ * (2.0 * u * v - u - v)
pdf = -theta_ * (uv - 1.0) / ((thetaMinus1_ * (u - v)) ** 2 - 2.0 * uv + 1.0) ** (3 / 2)

print("dpdf/dtheta=", sp.simplify(sp.diff(pdf, theta_)))

cdf = (
    1.0
    + thetaMinus1_ * (u + v)
    - ((1.0 + thetaMinus1_ * (u + v)) ** 2 - 4.0 * u * v * theta_ * thetaMinus1_) ** 0.5
) / (2.0 * thetaMinus1_)

print("dcdf/dtheta=", sp.simplify(sp.diff(cdf, theta_)))



#! /usr/bin/env python
"""
Consider a model exactly linear with respect to the parameters.
In this case, the LLSQ calibration performs as good as it can.
Considers a case without observed inputs.
"""

import openturns as ot
import openturns.testing as ott

ot.TESTPREAMBLE()
ot.PlatformInfo.SetNumericalPrecision(5)

ot.RandomGenerator.SetSeed(0)


# This model is linear in (a, b, c) and identifiable.
# Derived from y = a + b * x + c * x^2 at x=[-1.0, -0.6, -0.2, 0.2, 0.6, 1.0]
g = ot.SymbolicFunction(
    ["a", "b", "c"],
    [
        "a +  -1.0  * b +  1.0  * c",
        "a +  -0.6  * b +  0.36  * c",
        "a +  -0.2  * b +  0.04  * c",
        "a +  0.2  * b +  0.04  * c",
        "a +  0.6  * b +  0.36  * c",
        "a +  1.0  * b +  1.0  * c",
    ],
)
inputDimension = g.getInputDimension()
outputDimension = g.getOutputDimension()

trueParameter = ot.Point([12.0, 7.0, -8])

parameterDimension = trueParameter.getDimension()

Theta1 = ot.Dirac(trueParameter[0])
Theta2 = ot.Dirac(trueParameter[1])
Theta3 = ot.Dirac(trueParameter[2])

inputRandomVector = ot.JointDistribution([Theta1, Theta2, Theta3])

candidate = ot.Point([8.0, 9.0, -6.0])

calibratedIndices = [0, 1, 2]
model = ot.ParametricFunction(g, calibratedIndices, candidate)

outputObservationNoiseSigma = 0.01
meanNoise = ot.Point(outputDimension)
covarianceNoise = ot.Point(outputDimension, outputObservationNoiseSigma)
R = ot.IdentityMatrix(outputDimension)
observationOutputNoise = ot.Normal(meanNoise, covarianceNoise, R)

size = 100
inputObservations = ot.Sample(size, 0)

# Generate exact outputs
inputSample = inputRandomVector.getSample(size)
outputStress = g(inputSample)
# Add noise
sampleNoiseH = observationOutputNoise.getSample(size)
outputObservations = outputStress + sampleNoiseH

priorCovariance = ot.CovarianceMatrix(inputDimension)
for i in range(inputDimension):
    priorCovariance[i, i] = 3.0 + (1.0 + i) * (1.0 + i)
    for j in range(i):
        priorCovariance[i, j] = 1.0 / (1.0 + i + j)
errorCovariance = ot.CovarianceMatrix(outputDimension)
for i in range(outputDimension):
    errorCovariance[i, i] = 2.0 + (1.0 + i) * (1.0 + i)
    for j in range(i):
        errorCovariance[i, j] = 1.0 / (1.0 + i + j)
globalErrorCovariance = ot.CovarianceMatrix(outputDimension * size)
for i in range(outputDimension * size):
    globalErrorCovariance[i, i] = 0.1 * (2.0 + (1.0 + i) * (1.0 + i))
    for j in range(i):
        globalErrorCovariance[i, j] = 0.1 / (1.0 + i + j)

methods = ["SVD", "QR", "Cholesky"]
for method in methods:
    print("method=", method)
    # 1. Check with local error covariance
    print("Local error covariance")
    algo = ot.GaussianLinearCalibration(
        model,
        inputObservations,
        outputObservations,
        candidate,
        priorCovariance,
        errorCovariance,
        method,
    )
    algo.run()
    calibrationResult = algo.getResult()

    # Analysis of the results
    # Maximum A Posteriori estimator
    parameterMAP = calibrationResult.getParameterMAP()
    print("MAP=", repr(parameterMAP))
    rtol = 0.0
    atol = 1.0
    ott.assert_almost_equal(parameterMAP, trueParameter, rtol, atol)

    # 2. Check with global error covariance
    print("Global error covariance")
    algo = ot.GaussianLinearCalibration(
        model,
        inputObservations,
        outputObservations,
        candidate,
        priorCovariance,
        globalErrorCovariance,
        method,
    )
    algo.run()
    calibrationResult = algo.getResult()

    # Analysis of the results
    # Maximum A Posteriori estimator
    parameterMAP = calibrationResult.getParameterMAP()
    print("MAP=", repr(parameterMAP))
    rtol = 0.0
    atol = 1.0
    ott.assert_almost_equal(parameterMAP, trueParameter, rtol, atol)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


class RVEC(ot.PythonRandomVector):
    def __init__(self):
        # https://github.com/uqfoundation/dill/issues/300
        # super(RVEC, self).__init__(2)
        ot.PythonRandomVector.__init__(self, 2)
        self.setDescription(["R", "S"])

    def getRealization(self):
        X = [ot.RandomGenerator.Generate(), 2 + ot.RandomGenerator.Generate()]
        return X

    def getSample(self, size):
        X = []
        for i in range(size):
            X.append([ot.RandomGenerator.Generate(), 2 + ot.RandomGenerator.Generate()])
        return X

    def getMean(self):
        return [0.5, 2.5]


rv = ot.RandomVector(RVEC())
print(rv)
print(rv.getRealization())
print(rv.getMean())

# save
study = ot.Study()
study.setStorageManager(ot.XMLStorageManager("pyrv.xml"))
study.add("rv", rv)
study.save()



import openturns as ot
from math import exp
from matplotlib import pyplot as plt
import openturns.viewer as otv


# Create the frequency grid
fmin = 0.1
df = 0.2
N = int((10.0 - fmin) / df)
myFrequencyGrid = ot.RegularGrid(fmin, df, N)

# Spectral function


def s(f):
    if f <= 5.0:
        return 1.0
    else:
        x = f - 5.0
        return exp(-2.0 * x * x)


# Create the collection of HermitianMatrix
myCollection = ot.HermitianMatrixCollection()
for k in range(N):
    frequency = myFrequencyGrid.getValue(k)
    matrix = ot.HermitianMatrix(1)
    matrix[0, 0] = s(frequency)
    myCollection.add(matrix)


# Create the spectral model
mySpectralModel = ot.UserDefinedSpectralModel(myFrequencyGrid, myCollection)

# Graph of the spectral function
x = ot.Sample(N, 2)
for k in range(N):
    frequency = myFrequencyGrid.getValue(k)
    x[k, 0] = frequency
    value = mySpectralModel(frequency)
    x[k, 1] = value[0, 0].real

# Create the curve of the spectral function
myCurve = ot.Curve(x, "UserSpectral")

# Create the graph
myGraph = ot.Graph(
    "User Defined Spectral Model", "Frequency", "Spectral density value", True
)
myGraph.add(myCurve)
myGraph.setLegendPosition("upper right")

fig = plt.figure(figsize=(10, 4))
cov_axis = fig.add_subplot(111)
view = otv.View(myGraph, figure=fig, axes=[cov_axis], add_legend=False)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

myFunc = ot.SymbolicFunction("x", "x^2")
tg = ot.RegularGrid(0.0, 0.2, 6)
mySpatialFunc = ot.ValueFunction(myFunc, tg)

print("mySpatialFunc=", mySpatialFunc)
# Get the input and output description
print("mySpatialFunc input description=", mySpatialFunc.getInputDescription())
print("mySpatialFunc output description=", mySpatialFunc.getOutputDescription())
# Get the input and output dimension, based on description
print("mySpatialFunc input dimension=", mySpatialFunc.getInputDimension())
print("mySpatialFunc output dimension=", mySpatialFunc.getOutputDimension())
# Create a TimeSeries
data = ot.Sample(tg.getN(), myFunc.getInputDimension())
for i in range(data.getSize()):
    for j in range(data.getDimension()):
        data[i, j] = i * data.getDimension() + j
ts = ot.TimeSeries(tg, data)
print("input time series=", ts)
print("output time series=", mySpatialFunc(ts))
# Get the number of calls
print("called ", mySpatialFunc.getCallsNumber(), " times")



"""
Estimate a probability using Line Sampling
==========================================
"""

# %%
# In this example, we estimate the probability that the output of a function exceeds a given threshold with the Line Sampling method.

# %%
import openturns as ot
import openturns.experimental as otexp
import openturns.viewer as otv
import matplotlib.pyplot as plt

# %%
# Define the limit state function and the random vector
# -----------------------------------------------------
dim = 2
g_twoBranch = ot.SymbolicFunction(
    ["x1", "x2"],
    [
        "min(4 + 0.1 * (x1 - x2)^2 - (x1 + x2) / sqrt(2), 4 + 0.1 * (x1 - x2)^2 + (x1 + x2) / sqrt(2))"
    ],
)
X = ot.RandomVector(ot.Normal(dim))

# %%
# Define the event
# ----------------
Y_twoBranch = ot.CompositeRandomVector(g_twoBranch, X)
threshold = 1.5
event_twoBranch = ot.ThresholdEvent(Y_twoBranch, ot.Less(), threshold)

# %%
# Run FORM approximation
# ----------------------
optimAlgo = ot.Cobyla()
optimAlgo.setStartingPoint(X.getMean())
algo = ot.FORM(optimAlgo, event_twoBranch)
algo.run()
resultFORM = algo.getResult()


# %%
# Run Line Sampling
# -----------------

# %%
# The design point will define the initial important direction
alpha_twoBranch = resultFORM.getStandardSpaceDesignPoint()

# %%
# Define the root solver
solver = ot.Brent(1e-3, 1e-3, 1e-3, 5)
rootStrategy = ot.MediumSafe(solver)

# %%
# Create the algorithm
algo = otexp.LineSampling(event_twoBranch, alpha_twoBranch, rootStrategy)
algo.setMaximumOuterSampling(1000)
algo.setMaximumCoefficientOfVariation(5e-2)
# disable adaptive important direction to make plots easier to interpret
algo.setAdaptiveImportantDirection(False)

# %%
# Save the important direction history, and especially all root points
algo.setStoreHistory(True)

# %%
# Run the simulation
algo.run()
result = algo.getResult()
print(result)

# %%
# Retrieve important directions
alphas = algo.getAlphaHistory()

# %%
# Retrieve root points and root values
rootPoints = algo.getRootPointsHistory()
rootValues = algo.getRootValuesHistory()


# %%


def drawLines(algo, n=10):
    """Draw sampling lines and the corresponding roots."""
    rootPoints = algo.getRootPointsHistory()
    alphas = algo.getAlphaHistory()
    n = min(n, len(rootPoints))
    for i in range(n):
        # there can be several roots per sample
        n_roots = len(rootPoints[i])
        alpha = alphas[i]
        for j in range(n_roots):
            if i + j == 0:
                print(f"n_roots={n_roots}")
            dp = rootPoints[i][j]
            # project design point on the hyperplane orthogonal to alpha to get origin point
            uPoint = dp
            uDot = uPoint.dot(alpha)
            uPoint = uPoint - alpha * uDot
            # draw segment origin -> design point
            plt.plot(
                (uPoint[0], dp[0]),
                (uPoint[1], dp[1]),
                color="blue",
                linestyle="--",
                linewidth=0.75,
            )
            # draw origin
            plt.plot(uPoint[0], uPoint[1], "ro", markersize=3)
            # draw design point
            plt.plot(dp[0], dp[1], "bo", markersize=3, zorder=3)


# %%


def drawLSDesign(algo):
    """Draw sampling lines, roots, and the limit state function."""
    dmin = [-4.0] * 2
    dmax = [4.0] * 2
    graph1 = g_twoBranch.draw(dmin, dmax)
    contour_g = graph1.getDrawable(0).getImplementation()
    contour_g.setColorBarPosition("")
    contour_g.setColorMap("gray")
    graph1.setDrawable(0, contour_g)
    view1 = otv.View(graph1, square_axes=True)
    # now draw the the limit state on same figure
    graph2 = g_twoBranch.draw(dmin, dmax)
    contour_g = graph2.getDrawable(0).getImplementation()
    contour_g.setLevels([threshold])
    contour_g.setColor("red")
    graph2.setDrawable(0, contour_g)
    plt.axline([-1, 1], [1, -1], linestyle="dotted", color="gray")
    drawLines(algo)
    graph2.setTitle("Line Sampling")
    _ = otv.View(graph2, figure=view1.getFigure())


# %%
# Plot the limit state, a few design points and their origin
drawLSDesign(algo)

# %%
# Now we disable the opposite direction search to try to save function evaluations.
# This practice is incorrect in this case, however, as the algorithm misses half the probability of the event.
ot.RandomGenerator.SetSeed(0)
algo.setSearchOppositeDirection(False)
algo.run()
result = algo.getResult()
print(result)

# %%
# We plot the limit state, a few design points and their origin.
# This time we see each origin sample point in the orthogonal hyperplane yields only one design point.
drawLSDesign(algo)

# %%
# Now re-enable adaptive important direction search
ot.RandomGenerator.SetSeed(0)
algo.setAdaptiveImportantDirection(True)
algo.setSearchOppositeDirection(True)
algo.run()
result = algo.getResult()
print(result)

# %%
# Inspect important directions (without duplicates from history)
unique_alphas = ot.Sample(0, dim)
for alpha in algo.getAlphaHistory():
    if len(unique_alphas) == 0 or alpha != unique_alphas[-1]:
        unique_alphas.add(alpha)
print("unique alphas:")
print(unique_alphas)

# %%
otv.View.ShowAll()



"""
Metamodel of a field function
=============================
"""

# %%
#
# In this example we create a metamodel of a field function following these steps:
#
# 1. Create a field model over an 1-d mesh ;
# 2. Create a Gaussian process ;
# 3. Karhunen-Loeve decomposition of a process with known covariance function ;
# 4. Karhunen-Loeve decomposition of a process with known trajectories ;
# 5. Fields projection ;
# 6. Functional chaos decomposition between the coefficients of the input and output processes ;
# 7. Build a metamodel of the whole field model ;
# 8. Validate the metamodel.
#

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# Create the input model.

print("Create the input process")
# Domain bound
a = 1
# Reference correlation length
b = 0.5
# Number of vertices in the mesh
N = 50
# Bandwidth of the smoothers
h = 0.05

mesh = ot.IntervalMesher([N - 1]).build(ot.Interval(-a, a))
covariance_X = ot.AbsoluteExponential([b])
process_X = ot.GaussianProcess(covariance_X, mesh)


# %%
# The next function plots the Karhunen-Loève modes.
def drawKL(scaledKL, KLev, mesh, title="Scaled KL modes"):
    graph_modes = scaledKL.drawMarginal()
    graph_modes.setTitle(title + " scaled KL modes")
    graph_modes.setXTitle("$x$")
    graph_modes.setYTitle(r"$\sqrt{\lambda_i}\phi_i$")
    data_ev = [[i, KLev[i]] for i in range(scaledKL.getSize())]
    graph_ev = ot.Graph()
    graph_ev.add(ot.Curve(data_ev))
    graph_ev.add(ot.Cloud(data_ev))
    graph_ev.setTitle(title + " KL eigenvalues")
    graph_ev.setXTitle("$k$")
    graph_ev.setYTitle(r"$\lambda_i$")
    graph_ev.setAxes(True)
    graph_ev.setGrid(True)
    graph_ev.setLogScale(2)
    bb = graph_ev.getBoundingBox()
    lower = bb.getLowerBound()
    lower[1] = 1.0e-7
    bb = ot.Interval(lower, bb.getUpperBound())
    graph_ev.setBoundingBox(bb)
    return graph_modes, graph_ev


# %%
# Compute the decomposition of the input process.
threshold = 0.01
algo_X = ot.KarhunenLoeveP1Algorithm(mesh, process_X.getCovarianceModel(), threshold)
algo_X.run()
result_X = algo_X.getResult()
phi_X = result_X.getScaledModesAsProcessSample()
lambda_X = result_X.getEigenvalues()

graph_modes_X, graph_ev_X = drawKL(phi_X, lambda_X, mesh, "X")
view = otv.View(graph_modes_X)


# %%
# Sample the input process.
print("Sample the input process")
size = 500
sample_X = process_X.getSample(size)

# %%
# Create a class to perform the convolution over a 1-d mesh.


class ConvolutionP1(ot.OpenTURNSPythonFieldFunction):
    def __init__(self, p, mesh):
        # 1 = input dimension, the dimension of the input field
        # 1 = output dimension, the dimension of the output field
        # 1 = mesh dimension
        super(ConvolutionP1, self).__init__(mesh, 1, mesh, 1)
        # Here we define some constants and we set-up the invariant part of the execution
        self.setInputDescription(["x"])
        self.setOutputDescription(["y"])
        vertices = mesh.getVertices()
        size = vertices.getSize()
        self.mat_W_ = ot.SquareMatrix(size)
        for i in range(size):
            x_minus_t = (vertices - vertices[i]) * (-1.0)
            values_w = p(x_minus_t)
            for j in range(size):
                self.mat_W_[i, j] = values_w[j, 0]
        G = mesh.computeP1Gram()
        self.mat_W_ = self.mat_W_ * G

    def _exec(self, X):
        point_X = [val[0] for val in X]
        values_Y = self.mat_W_ * point_X
        return [[v] for v in values_Y]


# %%
# Create the convolution function.
p = ot.SymbolicFunction("x", "exp(-(x/" + str(h) + ")^2)")
myConvolution = ot.FieldFunction(ConvolutionP1(p, mesh))

# %%
# Sample the output process.
sample_Y = myConvolution(sample_X)

# %%
# Compute the decomposition of the output process.
algo_Y = ot.KarhunenLoeveSVDAlgorithm(sample_Y, threshold)
algo_Y.run()
result_Y = algo_Y.getResult()
phi_Y = result_Y.getScaledModesAsProcessSample()
lambda_Y = result_Y.getEigenvalues()
graph_modes_Y, graph_ev_Y = drawKL(phi_Y, lambda_Y, mesh, "Y")
view = otv.View(graph_modes_Y)

# %%
# Compare eigenvalues of X and Y.
graph_ev_X.add(graph_ev_Y)
graph_ev_X.setTitle("Input/output eigenvalues comparison")
graph_ev_X.setYTitle(r"$\lambda_X, \lambda_Y$")
graph_ev_X.setColors(["blue", "blue", "red", "red"])
graph_ev_X.setLegends([r"$\lambda_X$", "", r"$\lambda_Y$", ""])
graph_ev_X.setLegendPosition("upper right")
view = otv.View(graph_ev_X)

# %%
# Perform the polynomial chaos expansion between Karhunen-Loève coefficients.
print("project sample_X")
sample_xi_X = result_X.project(sample_X)

print("project sample_Y")
sample_xi_Y = result_Y.project(sample_Y)

print("Compute PCE between coefficients")
degree = 1
dimension_xi_X = sample_xi_X.getDimension()
dimension_xi_Y = sample_xi_Y.getDimension()
enumerateFunction = ot.LinearEnumerateFunction(dimension_xi_X)
basis = ot.OrthogonalProductPolynomialFactory(
    [ot.HermiteFactory()] * dimension_xi_X, enumerateFunction
)
basisSize = enumerateFunction.getStrataCumulatedCardinal(degree)
adaptive = ot.FixedStrategy(basis, basisSize)
projection = ot.LeastSquaresStrategy(
    ot.LeastSquaresMetaModelSelectionFactory(ot.LARS(), ot.CorrectedLeaveOneOut())
)
ot.ResourceMap.SetAsScalar("LeastSquaresMetaModelSelection-ErrorThreshold", 1.0e-7)
algo_chaos = ot.FunctionalChaosAlgorithm(
    sample_xi_X, sample_xi_Y, basis.getMeasure(), adaptive, projection
)
algo_chaos.run()
result_chaos = algo_chaos.getResult()
meta_model = result_chaos.getMetaModel()
print(
    "myConvolution=",
    myConvolution.getInputDimension(),
    "->",
    myConvolution.getOutputDimension(),
)
preprocessing = ot.KarhunenLoeveProjection(result_X)
print(
    "preprocessing=",
    preprocessing.getInputDimension(),
    "->",
    preprocessing.getOutputDimension(),
)
print(
    "meta_model=", meta_model.getInputDimension(), "->", meta_model.getOutputDimension()
)
postprocessing = ot.KarhunenLoeveLifting(result_Y)
print(
    "postprocessing=",
    postprocessing.getInputDimension(),
    "->",
    postprocessing.getOutputDimension(),
)
meta_model_field = ot.FieldToFieldConnection(
    postprocessing, ot.FieldToPointConnection(meta_model, preprocessing)
)


# %%
# Validate the metamodel.
iMax = 10
sample_X_validation = process_X.getSample(iMax)
sample_Y_validation = myConvolution(sample_X_validation)

graph_sample_Y_validation = sample_Y_validation.drawMarginal(0)
sample_Y_hat = meta_model_field(sample_X_validation)
graph = sample_Y_hat.drawMarginal(0)
for i in range(iMax):
    dr = graph.getDrawable(i)
    dr.setLineStyle("dashed")
    graph_sample_Y_validation.add(dr)
graph_sample_Y_validation.setTitle(r"Comparison $Y_i$ and $\tilde{Y}_i$")
graph_sample_Y_validation.setXTitle(r"$t$")
graph_sample_Y_validation.setYTitle(r"$Y$, $\tilde{Y}$")
view = otv.View(graph_sample_Y_validation)

# %%
graph_sample_X = sample_X_validation.drawMarginal(0)
graph_sample_X.setTitle(r"Trajectory $X$")
graph_sample_X.setXTitle(r"$t$")
graph_sample_X.setYTitle(r"$X$")
view = otv.View(graph_sample_X)

# %%
graph_sample_Y = sample_Y_validation.drawMarginal(0)
graph_sample_Y.setTitle(r"Trajectory $Y$")
graph_sample_Y.setXTitle(r"$t$")
graph_sample_Y.setYTitle(r"$Y$")
view = otv.View(graph_sample_Y)

# %%
# Display all figures
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()
ot.PlatformInfo.SetNumericalPrecision(4)

# linear
levelFunction = ot.SymbolicFunction(["x1", "x2", "x3", "x4"], ["x1+2*x2-3*x3+4*x4"])
algo = ot.Cobyla(ot.NearestPointProblem(levelFunction, 3.0))
algo.setStartingPoint([0.0] * 4)
print("algo=", algo)
algo.run()
result = algo.getResult()
print("x^=", result.getOptimalPoint())
print("f(x^)=", result.getOptimalValue())
print("lambda^=", result.computeLagrangeMultipliers())

# non-linear
levelFunction = ot.SymbolicFunction(
    ["x1", "x2", "x3", "x4"], ["x1*cos(x1)+2*x2*x3-3*x3+4*x3*x4"]
)
algo = ot.Cobyla(ot.NearestPointProblem(levelFunction, 3.0))
algo.setStartingPoint([0.0] * 4)
algo.setMaximumCallsNumber(400)
algo.setMaximumAbsoluteError(1.0e-10)
algo.setMaximumRelativeError(1.0e-10)
algo.setMaximumResidualError(1.0e-10)
algo.setMaximumConstraintError(1.0e-10)
algo.run()
result = algo.getResult()
print("x^=", result.getOptimalPoint())
print("f(x^)=", result.getOptimalValue())
print("lambda^=", result.computeLagrangeMultipliers())

# bounds
linear = ot.SymbolicFunction(["x1", "x2", "x3", "x4"], ["x1+2*x2-3*x3+4*x4"])

dim = 4

bounds = ot.Interval([-3.0] * dim, [5.0] * dim)

for minimization in [True, False]:
    problem = ot.OptimizationProblem(linear, ot.Function(), ot.Function(), bounds)
    problem.setMinimization(minimization)
    algo = ot.Cobyla(problem)
    algo.setMaximumCallsNumber(150)
    algo.setStartingPoint([0.0] * dim)
    print("algo=", algo)
    algo.run()
    result = algo.getResult()
    print("x^=", result.getOptimalPoint())
    print("f(x^)=", result.getOptimalValue())
    print("lambda^=", result.computeLagrangeMultipliers())

# empty problem
algo = ot.Cobyla()
try:
    algo.run()
except Exception:
    print("OK")



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

# Create a Box Cox transformation
lambdas = [0.0, 0.1, 1.0, 1.5]

graph = ot.Graph("Box Cox transformations", "x", "y")
for i in range(len(lambdas)):
    boxCoxT = ot.BoxCoxTransform(lambdas[i])
    graph.add(boxCoxT.draw(0.1, 2.1))

graph.setLegends(["lambda = " + str(lam) for lam in lambdas])
graph.setLegendPosition("lower right")

fig = plt.figure(figsize=(8, 4))
axis = fig.add_subplot(111)
axis.set_xlim(auto=True)
otv.View(graph, figure=fig, axes=[axis], add_legend=True)



#! /usr/bin/env python

import openturns as ot
from openturns.testing import assert_almost_equal
from openturns.usecases import branin_function


ot.TESTPREAMBLE()
ot.PlatformInfo.SetNumericalPrecision(5)

"""
Test the import of the BraninModel data class.
"""
bm = branin_function.BraninModel()

# test parameters
assert_almost_equal(bm.dim, 2, 1e-12)
assert_almost_equal(bm.trueNoiseFunction, 0.1, 1e-12)
assert_almost_equal(bm.lowerbound, ot.Point([0.0] * bm.dim), 1e-12)
assert_almost_equal(bm.upperbound, ot.Point([1.0] * bm.dim), 1e-12)

# test minima
assert_almost_equal(bm.xexact1, ot.Point([0.123895, 0.818329]), 1e-12)
assert_almost_equal(bm.xexact2, ot.Point([0.542773, 0.151666]), 1e-12)
assert_almost_equal(bm.xexact3, ot.Point([0.961652, 0.165000]), 1e-12)



import openturns as ot
import openturns.viewer as otv

g = ot.SymbolicFunction(["x"], ["sin(x)"])
f = ot.SymbolicFunction(["y"], ["abs(y)"])
composed = ot.ComposedFunction(f, g)

graph = composed.draw(0.0, 10.0)
graph.setTitle("y=abs(sin(x))")
otv.View(graph, figure_kw={"figsize": (8, 4)}, add_legend=True).ShowAll()



import openturns as ot
import openturns.viewer as otv

graph = ot.Graph("BarPlot", "X", "Y")
graph.setLegendPosition("upper right")
origin = 0.0
data = [[1.0, 2.0], [3.0, 4.0], [2.0, 3.0]]
barPlot = ot.BarPlot(data, origin, "orange", "shaded", "dashed", "label")
graph.add(barPlot)

otv.View(graph, figure_kw={"figsize": (6, 4)})



"""
Create a mesh
=============
"""

# %%
import openturns as ot
import openturns.viewer as otv
import math as m

# %%
# Creation of a regular grid
# --------------------------
#
# In this first part we demonstrate how to create a regular grid. Note that a regular grid is a particular mesh of :math:`\mathcal{D}=[0,T] \in \mathbb{R}`.
#
# Here we assume it represents the time :math:`t` as it is often the case, but it can represent any physical quantity.
#
# A regular time grid is a regular discretization of the interval :math:`[0, T] \in \mathbb{R}` into :math:`N` points, noted :math:`(t_0, \dots, t_{N-1})`.
#
# The time grid can be defined using :math:`(t_{Min}, \Delta t, N)` where :math:`N` is the number of points in the time grid.
# :math:`\Delta t` the time step between two consecutive instants and :math:`t_0 = t_{Min}`.
# Then, :math:`t_k = t_{Min} + k \Delta t` and :math:`t_{Max} = t_{Min} +  (N-1) \Delta t`.
#
#
# Consider :math:`X: \Omega \times \mathcal{D} \rightarrow \mathbb{R}^d` a multivariate stochastic process of dimension :math:`d`,
# where :math:`n=1`, :math:`\mathcal{D}=[0,T]` and :math:`t\in \mathcal{D}` is interpreted as a time stamp.
# Then the mesh associated to the process :math:`X` is a (regular) time grid.


# %%
# We define a time grid from a starting time :math:`tMin`, a time step :math:`tStep` and a number of time steps :math:`n`.
tMin = 0.0
tStep = 0.1
n = 10
time_grid = ot.RegularGrid(tMin, tStep, n)

# %%
# We get the first and the last instants, the step and the number of points :
start = time_grid.getStart()
step = time_grid.getStep()
grid_size = time_grid.getN()
end = time_grid.getEnd()
print("start=", start, "step=", step, "grid_size=", grid_size, "end=", end)

# %%
# We draw the grid.
time_grid.setName("time")
graph = time_grid.draw()
graph.setTitle("Time grid")
graph.setXTitle("t")
graph.setYTitle("")
view = otv.View(graph)


# %%
# Creation of a mesh
# ------------------

# %%
# In this paragraph we create a mesh :math:`\mathcal{M}` associated to a domain :math:`\mathcal{D} \in \mathbb{R}^n`.
#
# A mesh is defined from vertices in :math:`\mathbb{R}^n` and a topology that connects the vertices: the simplices.
# The simplex :math:`Indices([i_1,\dots, i_{n+1}])` relies the vertices of index :math:`(i_1,\dots, i_{n+1})` in :math:`\mathbb{N}^n`.
# In dimension 1, a simplex is an interval :math:`Indices([i_1,i_2])`; in dimension 2, it is a triangle :math:`Indices([i_1,i_2, i_3])`.
#
# The library enables to easily create a mesh which is a box of dimension :math:`d=1` or :math:`d=2` regularly meshed in all its directions, thanks to the object IntervalMesher.
#
# Consider :math:`X: \Omega \times \mathcal{D} \rightarrow \mathbb{R}^d` a multivariate stochastic process of dimension :math:`d`, where :math:`\mathcal{D} \in \mathbb{R}^n`.
# The mesh :math:`\mathcal{M}` is a discretization of the domain :math:`\mathcal{D}`.

# %%
# A  one dimensional mesh is created and represented by :
vertices = [[0.5], [1.5], [2.1], [2.7]]
simplicies = [[0, 1], [1, 2], [2, 3]]
mesh1D = ot.Mesh(vertices, simplicies)
graph1 = mesh1D.draw()
graph1.setTitle("One dimensional mesh")
view = otv.View(graph1)

# %%
# We define a bidimensional mesh :
vertices = [[0.0, 0.0], [1.0, 0.0], [1.0, 1.0], [1.5, 1.0], [2.0, 1.5], [0.5, 1.5]]
simplicies = [[0, 1, 2], [1, 2, 3], [2, 3, 4], [2, 4, 5], [0, 2, 5]]
mesh2D = ot.Mesh(vertices, simplicies)
graph2 = mesh2D.draw()
graph2.setTitle("Bidimensional mesh")
graph2.setLegendPosition("lower right")
view = otv.View(graph2)

# %%
# We can also define a mesh which is a regularly meshed box in dimension 1 or 2.
# We define the number of intervals in each direction of the box :
myIndices = [5, 10]
myMesher = ot.IntervalMesher(myIndices)

# %%
# We then create the mesh of the box :math:`[0, 2] \times [0, 4]` :
lowerBound = [0.0, 0.0]
upperBound = [2.0, 4.0]
myInterval = ot.Interval(lowerBound, upperBound)
myMeshBox = myMesher.build(myInterval)
mygraph3 = myMeshBox.draw()
mygraph3.setTitle("Bidimensional mesh on a box")
view = otv.View(mygraph3)

# %%
# It is possible to perform a transformation on a regularly meshed box.
myIndices = [20, 20]
mesher = ot.IntervalMesher(myIndices)
# r in [1., 2.] and theta in (0., pi]
lowerBound2 = [1.0, 0.0]
upperBound2 = [2.0, m.pi]
myInterval = ot.Interval(lowerBound2, upperBound2)
meshBox2 = mesher.build(myInterval)

# %%
# We define the mapping function and draw the transformation :
f = ot.SymbolicFunction(["r", "theta"], ["r*cos(theta)", "r*sin(theta)"])
oldVertices = meshBox2.getVertices()
newVertices = f(oldVertices)
meshBox2.setVertices(newVertices)
graphMappedBox = meshBox2.draw()
graphMappedBox.setTitle("Mapped box mesh")
view = otv.View(graphMappedBox)


# %%
# Finally we create a mesh of a heart in dimension 2.
# sphinx_gallery_thumbnail_number = 6
def meshHeart(ntheta, nr):
    # First, build the nodes
    nodes = ot.Sample(0, 2)
    nodes.add([0.0, 0.0])
    for j in range(ntheta):
        theta = (m.pi * j) / ntheta
        if abs(theta - 0.5 * m.pi) < 1e-10:
            rho = 2.0
        elif (abs(theta) < 1e-10) or (abs(theta - m.pi) < 1e-10):
            rho = 0.0
        else:
            absTanTheta = abs(m.tan(theta))
            rho = absTanTheta ** (1.0 / absTanTheta) + m.sin(theta)
        cosTheta = m.cos(theta)
        sinTheta = m.sin(theta)
        for k in range(nr):
            tau = (k + 1.0) / nr
            r = rho * tau
            nodes.add([r * cosTheta, r * sinTheta - tau])
    # Second, build the triangles
    triangles = []
    # First heart
    for j in range(ntheta):
        triangles.append([0, 1 + j * nr, 1 + ((j + 1) % ntheta) * nr])
    # Other hearts
    for j in range(ntheta):
        for k in range(nr - 1):
            i0 = k + 1 + j * nr
            i1 = k + 2 + j * nr
            i2 = k + 2 + ((j + 1) % ntheta) * nr
            i3 = k + 1 + ((j + 1) % ntheta) * nr
            triangles.append([i0, i1, i2 % (nr * ntheta)])
            triangles.append([i0, i2, i3 % (nr * ntheta)])
    return ot.Mesh(nodes, triangles)


mesh4 = meshHeart(48, 16)
graphMesh = mesh4.draw()
graphMesh.setTitle("Bidimensional mesh")
graphMesh.setLegendPosition("")
view = otv.View(graphMesh)

# %%
# Display all figures
otv.View.ShowAll()



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

f1 = ot.SymbolicFunction(["t"], ["sin(t)"])
f2 = ot.SymbolicFunction(["t"], ["cos(t)*cos(t)"])
myBasis = ot.Basis([f1, f2])
coefDis = ot.Normal([2] * 2, [5] * 2, ot.CorrelationMatrix(2))
myTG = ot.RegularGrid(0.0, 0.1, 250)

myFBP = ot.FunctionalBasisProcess(coefDis, myBasis, myTG)

TS = myFBP.getRealization()


graph = TS.draw()
graph.add(myFBP.getRealization().draw())
graph.add(myFBP.getRealization().draw())
graph.setTitle("Functional Basis Process")

fig = plt.figure(figsize=(10, 4))
fbp_axis = fig.add_subplot(111)
view = otv.View(graph, figure=fig, axes=[fbp_axis], add_legend=False)



#! /usr/bin/env python

import openturns as ot

# Polynomial factories
factoryCollection = [ot.LaguerreFactory(2.5), ot.LegendreFactory(), ot.HermiteFactory()]
dim = len(factoryCollection)
basisFactory = ot.OrthogonalProductPolynomialFactory(factoryCollection)
print("basisFactory=")
print(basisFactory)
print(basisFactory.__repr_markdown__())
print(basisFactory._repr_html_())

basis = ot.OrthogonalBasis(basisFactory)
print("basis=")
print(basis)

x = [0.5] * dim
for i in range(10):
    f = basis.build(i)
    print("i=", i, "f(X)=", f(x))

# Using multi-indices
enum = basis.getEnumerateFunction()
for i in range(10):
    indices = enum(i)
    f = basis.build(indices)
    print("indices=", indices, "f(X)=", f(x))

# Other factories
factoryCollection = [
    ot.OrthogonalUniVariatePolynomialFunctionFactory(ot.LaguerreFactory(2.5)),
    ot.HaarWaveletFactory(),
    ot.FourierSeriesFactory(),
]
dim = len(factoryCollection)
basisFactory = ot.OrthogonalProductFunctionFactory(factoryCollection)
basis = ot.OrthogonalBasis(basisFactory)
print("basis=", basis)
x = [0.5] * dim
for i in range(10):
    f = basis.build(i)
    print("i=", i, "f(X)=", f(x))

# Using multi-indices
enum = basis.getEnumerateFunction()
for i in range(10):
    indices = enum(i)
    f = basis.build(indices)
    print("indices=", indices, "f(X)=", f(x))

# Polynomial factories using a collection of distributions
distributionCollection = [
    ot.Normal(),
    ot.TruncatedDistribution(ot.Normal(2.0, 1.5), 1.0, 4.0),
    ot.Uniform(),
]
basisFactory = ot.OrthogonalProductPolynomialFactory(distributionCollection)
print("basisFactory=")
print(basisFactory)
print(basisFactory.__repr_markdown__())
print(basisFactory._repr_html_())



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

# Create the time grid
# In the context of the spectral estimate or Fourier transform use,
# we use data blocs with size of form 2^p
tMin = 0.0
timeStep = 0.1
size = pow(2, 12)
myTimeGrid = ot.RegularGrid(tMin, timeStep, size)

# We fix the parameter of the Cauchy model
amplitude = [5]
scale = [3]
model = ot.CauchyModel(scale, amplitude)
gp = ot.SpectralGaussianProcess(model, myTimeGrid)

# Get a time series or a sample of time series
# myTimeSeries = gp.getRealization()
mySample = gp.getSample(1000)

mySegmentNumber = 10
myOverlapSize = 0.3

# Build a spectral model factory
myFactory = ot.WelchFactory(ot.Hann(), mySegmentNumber, myOverlapSize)

# Estimation on a TimeSeries or on a ProcessSample
# myEstimatedModel_TS = myFactory.build(myTimeSeries)
myEstimatedModel_PS = myFactory.buildAsUserDefinedSpectralModel(mySample)

# Change the filtering window
myFactory.setFilteringWindows(ot.Hamming())

# Get the FFT algorithm
myFFT = myFactory.getFFTAlgorithm()

# Get the frequencyGrid
frequencyGrid = myEstimatedModel_PS.getFrequencyGrid()

# With the model, we want to compare values
# We compare values computed with theoritical values
plotSample = ot.Sample(frequencyGrid.getN(), 3)

# Loop of comparison ==> data are saved in plotSample
for k in range(frequencyGrid.getN()):
    freq = frequencyGrid.getStart() + k * frequencyGrid.getStep()
    plotSample[k, 0] = freq
    plotSample[k, 1] = abs(myEstimatedModel_PS(freq)[0, 0])
    plotSample[k, 2] = abs(model(freq)[0, 0])


# Graph section
# We build 2 curves
# each one is function of frequency values
ind = ot.Indices(2)
ind.fill()

# Some cosmetics : labels, legend position, ...
graph = ot.Graph(
    "Spectral model estimation",
    "Frequency",
    "Spectral density function",
    True,
    "upper right",
    1.0,
    ot.GraphImplementation.LOGY,
)

# The first curve is the estimate density as function of frequency
curve1 = ot.Curve(plotSample.getMarginal(ind))
curve1.setColor("blue")
curve1.setLegend("estimate model")

# The second curve is the theoritical density as function of frequency
ind[1] = 2
curve2 = ot.Curve(plotSample.getMarginal(ind))
curve2.setColor("red")
curve2.setLegend("Cauchy model")

graph.add(curve1)
graph.add(curve2)

fig = plt.figure(figsize=(10, 4))
graph_axis = fig.add_subplot(111)
view = otv.View(graph, figure=fig, axes=[graph_axis], add_legend=False)



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

a = 4.0
covarianceModel = ot.FractionalBrownianMotionModel(1.0, 1.0, 0.25)
cov_graph = covarianceModel.draw(0, 0, -a, a, 512)
cov_graph.setTitle(str(covarianceModel))
fig = plt.figure(figsize=(10, 4))
cov_axis = fig.add_subplot(111)
otv.View(cov_graph, figure=fig, axes=[cov_axis], add_legend=False, square_axes=True)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


def sobol(indice, ai):
    val = 1.0
    for i in range(indice.getSize()):
        val = val * 1.0 / (3.0 * (1.0 + ai[indice[i]]) ** 2.0)
    return val


# Problem parameters
dimension = 3

# Create the Sobol function
# Reference analytical values
meanTh = 1.0
a = ot.Point(dimension)
inputVariables = ot.Description(dimension)
formula = ot.Description(1)
formula[0] = "1.0"
covTh = 1.0
for i in range(dimension):
    a[i] = 0.5 * i
    covTh = covTh * (1.0 + 1.0 / (3.0 * (1.0 + a[i]) ** 2))
    inputVariables[i] = "xi" + str(i)
    formula[0] = (
        formula[0]
        + " * ((abs(4.0 * xi"
        + str(i)
        + " - 2.0) + "
        + str(a[i])
        + ") / (1.0 + "
        + str(a[i])
        + "))"
    )
covTh = covTh - 1.0
model = ot.SymbolicFunction(inputVariables, formula)

# Create the input distribution
marginals = [ot.Uniform(0.0, 1.0)] * dimension
distribution = ot.JointDistribution(marginals)

# Create the orthogonal basis
enumerateFunction = ot.LinearEnumerateFunction(dimension)
productBasis = ot.OrthogonalProductPolynomialFactory(marginals)

# Create the adaptive strategy
# We can choose amongst several strategies
# First, the most efficient (but more complex!) strategy
listAdaptiveStrategy = list()
degree = 4
indexMax = enumerateFunction.getStrataCumulatedCardinal(degree)
basisDimension = enumerateFunction.getStrataCumulatedCardinal(degree // 2)
threshold = 1.0e-6
listAdaptiveStrategy.append(
    ot.CleaningStrategy(productBasis, indexMax, basisDimension, threshold)
)
# Second, the most used (and most basic!) strategy
listAdaptiveStrategy.append(
    ot.FixedStrategy(
        productBasis, enumerateFunction.getBasisSizeFromTotalDegree(degree)
    )
)

for adaptiveStrategyIndex in range(len(listAdaptiveStrategy)):
    adaptiveStrategy = listAdaptiveStrategy[adaptiveStrategyIndex]
    # Create the projection strategy
    samplingSize = 250
    listExperiment = list()
    # Monte Carlo sampling
    listExperiment.append(ot.MonteCarloExperiment(distribution, samplingSize))
    # LHS sampling
    listExperiment.append(ot.LHSExperiment(distribution, samplingSize))
    # Low Discrepancy sequence
    listExperiment.append(
        ot.LowDiscrepancyExperiment(ot.SobolSequence(), distribution, samplingSize)
    )
    for experiment in listExperiment:
        # Create the polynomial chaos algorithm
        maximumResidual = 1.0e-10
        ot.RandomGenerator.SetSeed(0)
        X = experiment.generate()
        Y = model(X)
        algo = ot.FunctionalChaosAlgorithm(X, Y, distribution, adaptiveStrategy)
        algo.setMaximumResidual(maximumResidual)
        algo.run()

        # Examine the results
        result = algo.getResult()
        print("###################################")
        print(algo.getAdaptiveStrategy())
        print(algo.getProjectionStrategy())
        # print "coefficients=", result.getCoefficients()
        print("isLeastSquares= ", result.isLeastSquares())
        assert result.isLeastSquares()
        print("involvesModelSelection= ", result.involvesModelSelection())
        modelSelectionReference = adaptiveStrategy.getClassName() == "CleaningStrategy"
        assert result.involvesModelSelection() == modelSelectionReference

        # Post-process the results
        vector = ot.FunctionalChaosRandomVector(result)
        mean = vector.getMean()[0]
        print("mean=%.8f" % mean, "absolute error=%.8f" % abs(mean - meanTh))
        variance = vector.getCovariance()[0, 0]
        sensitivity = ot.FunctionalChaosSobolIndices(result)
        print("variance=%.8f" % variance, "absolute error=%.8f" % abs(variance - covTh))
        indices = ot.Indices(1)
        for i in range(dimension):
            indices[0] = i
            value = sensitivity.getSobolIndex(i)
            print(
                "Sobol index",
                i,
                "= %.8f" % value,
                "absolute error=%.8f" % abs(value - sobol(indices, a) / covTh),
            )
        indices = ot.Indices(2)
        k = 0
        for i in range(dimension):
            indices[0] = i
            for j in range(i + 1, dimension):
                indices[1] = j
                value = sensitivity.getSobolIndex(indices)
                print(
                    "Sobol index",
                    indices,
                    "=%.8f" % value,
                    "absolute error=%.8f" % abs(value - sobol(indices, a) / covTh),
                )
                k = k + 1
        indices = ot.Indices([0, 1, 2])
        value = sensitivity.getSobolIndex(indices)
        print(
            "Sobol index",
            indices,
            "=%.8f" % value,
            "absolute error=%.8f" % abs(value - sobol(indices, a) / covTh),
        )



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


# Reduce the precision output as the estimation is based on a lazy
# optimizer
ot.PlatformInfo.SetNumericalPrecision(4)

# ARMA(p, q)
p = 1
q = 1

# ARMACoefficients initializing
arCoefficients = ot.Point(p, 0.80)
maCoefficients = ot.Point(q, 0.50)

# ARMA creation
myARMA = ot.ARMA(
    ot.ARMACoefficients(arCoefficients),
    ot.ARMACoefficients(maCoefficients),
    ot.WhiteNoise(ot.Normal(0.0, 0.05)),
)
myARMA.setTimeGrid(ot.RegularGrid(0.0, 0.1, 256))
print("myARMA process=", myARMA)

# Create a realization
timeSeries = myARMA.getRealization()

# Create a sample
sample = myARMA.getSample(100)

# First, build an ARMA based on a given order using the WhittleFactory
factory = ot.WhittleFactory(p, q)
print("factory=", factory)
print("factory as an ARMA factory=", ot.ARMAFactory(factory))
result, informationCriteria = factory.buildWithCriteria(ot.TimeSeries(timeSeries))
# print "Estimated ARMA=", result
# print "Information criteria=", informationCriteria
result2, informationCriteria = factory.buildWithCriteria(sample)
# print "Estimated ARMA=", result2
# print "Information criteria=", informationCriteria

# Second, build the best ARMA based on a given range of order using the
# WhittleFactory
pIndices = ot.Indices(p + 1)
pIndices.fill()
qIndices = ot.Indices(q + 1)
qIndices.fill()
factory = ot.WhittleFactory(pIndices, qIndices)
print("factory=", factory)
result, informationCriteria = factory.buildWithCriteria(ot.TimeSeries(timeSeries))
# print "Estimated ARMA=", result
# print "Information criteria=", informationCriteria
# print "History=", factory.getHistory()
result2, informationCriteria = factory.buildWithCriteria(sample)
# print "Estimated ARMA=", result2
# print "Information criteria=", informationCriteria
history = ot.WhittleFactoryStateCollection(factory.getHistory())
firstTestes = ot.WhittleFactoryState(history[0])
# print "History=", factory.getHistory()



import openturns as ot
import unittest
import openturns.testing as ott


class CheckKarhunenLoeveReduce(unittest.TestCase):
    def test_ZeroMean(self):
        # Create the KL result
        numberOfVertices = 10
        interval = ot.Interval(-1.0, 1.0)
        mesh = ot.IntervalMesher([numberOfVertices - 1]).build(interval)
        covariance = ot.SquaredExponential()
        process = ot.GaussianProcess(covariance, mesh)
        sampleSize = 10
        processSample = process.getSample(sampleSize)
        threshold = 0.0
        algo = ot.KarhunenLoeveSVDAlgorithm(processSample, threshold)
        algo.run()
        klresult = algo.getResult()
        # Create the KL reduction
        processSample.computeMean()
        klreduce = ot.KarhunenLoeveReduction(klresult)
        # Generate a trajectory and reduce it
        field = process.getRealization()
        values = field.getValues()
        reducedValues = klreduce(values)
        ott.assert_almost_equal(values, reducedValues)

    def test_NonZeroMean(self):
        # Create the KL result
        numberOfVertices = 10
        interval = ot.Interval(-1.0, 1.0)
        mesh = ot.IntervalMesher([numberOfVertices - 1]).build(interval)
        covariance = ot.SquaredExponential()
        zeroProcess = ot.GaussianProcess(covariance, mesh)
        # Define a trend function
        f = ot.SymbolicFunction(["t"], ["30 * t"])
        fTrend = ot.TrendTransform(f, mesh)
        # Add it to the process
        process = ot.CompositeProcess(fTrend, zeroProcess)
        # Sample
        sampleSize = 100
        processSample = process.getSample(sampleSize)
        threshold = 0.0
        algo = ot.KarhunenLoeveSVDAlgorithm(processSample, threshold)
        algo.run()
        klresult = algo.getResult()
        # Create the KL reduction
        processSample.computeMean()
        klreduce = ot.KarhunenLoeveReduction(klresult)
        # Generate a trajectory and reduce it
        field = process.getRealization()
        values = field.getValues()
        reducedValues = klreduce(values)
        ott.assert_almost_equal(values, reducedValues)

    def test_trend(self):
        N = 100
        M = 1000
        P = 10
        mean = ot.SymbolicFunction("x", "sign(x)")
        cov = ot.SquaredExponential([1.0], [0.1])
        mesh = ot.IntervalMesher([N]).build(ot.Interval(-2.0, 2.0))
        process = ot.GaussianProcess(ot.TrendTransform(mean, mesh), cov, mesh)
        sample = process.getSample(M)
        algo = ot.KarhunenLoeveSVDAlgorithm(sample, 1e-6)
        algo.run()
        result = algo.getResult()
        trend = ot.TrendTransform(ot.P1LagrangeEvaluation(sample.computeMean()), mesh)
        sample2 = process.getSample(P)
        sample2.setName("reduction of sign(x) w/o trend")
        reduced1 = ot.KarhunenLoeveReduction(result)(sample2)
        reduced2 = ot.KarhunenLoeveReduction(result, trend)(sample2)
        g = sample2.drawMarginal(0)
        g.setColors(["red"])
        g1 = reduced1.drawMarginal(0)
        g1.setColors(["blue"])
        drs = g1.getDrawables()
        for i, d in enumerate(drs):
            d.setLineStyle("dashed")
            drs[i] = d
        g1.setDrawables(drs)
        g.add(g1)
        g2 = reduced2.drawMarginal(0)
        g2.setColors(["green"])
        drs = g2.getDrawables()
        for i, d in enumerate(drs):
            d.setLineStyle("dotted")
            drs[i] = d
        g2.setDrawables(drs)
        g.add(g2)
        if 0:
            from openturns.viewer import View

            View(g).save("reduction.png")


if __name__ == "__main__":
    unittest.main()



"""
Visualize pairs
===============
"""

# %%
# In this example we are going to a visualize a multidimensional sample by drawing each pair of 2-d marginal clouds.

# %%
import openturns as ot
import openturns.viewer as otv

# %%
# Create data to visualize
R = ot.CorrelationMatrix(3)
for i in range(1, 3):
    R[i, i - 1] = -0.4
distribution = ot.Normal([0.0] * 3, [5.0] * 3, R)
sample = distribution.getSample(200)

# %%
# Create cloud pairs graphs
graph = ot.VisualTest.DrawPairs(sample)
view = otv.View(graph)

# %%
# Display all graphs
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

ot.ResourceMap.Set("SymbolicParser-Backend", "ExprTk")
evaluation = ot.SymbolicEvaluation(["x0", "x1", "x2"], ["y0"], ["x0^2+2*x1+3*x2^3"])
gradient = ot.SymbolicGradient(evaluation)
point = [-1.0, 4.0, -4.0]
print("gradient=", gradient)
print("value at point", point, "=", gradient.gradient(point))
# Check for symbolic differentiation with constants
f = ot.SymbolicEvaluation(["x"], ["y"], ["e_*x^2+x*_e+cos(pi_*x)+sin(x*pi_)"])
print("symbolic derivative=", ot.SymbolicGradient(f).getFormula(0, 0))



"""
Estimate a scalar ARMA process
==============================
"""

# %%
# The objective here is to estimate an ARMA model from a scalar stationary
# time series using the Whittle estimator and a centered Normal white
# noise.
#
# The data can be a unique time series or several time series collected in
# a process sample.
#
# If the user specifies the order :math:`(p,q)`, the library fits a model
# :math:`ARMA(p,q)` to the data by estimating the coefficients
# :math:`(a_1, \dots, a_p), (b_1, \dots, b_q)` and the variance :math:`\sigma` of the
# white noise.
#
# If the User specifies a range of orders :math:`(p,q) \in Ind_p \times Ind_q`,
# where :math:`Ind_p = [p_1, p_2]` and :math:`Ind_q = [q_1, q_2]`, We find the *best*
# model :math:`ARMA(p,q)` that fits to the data and estimates the corresponding
# coefficients.
#
# We proceed as follows:
#
# -   the object :class:`~openturns.WhittleFactory` is created with either a specified order
#     :math:`(p,q)` or a range :math:`Ind_p \times Ind_q`. By default, the Welch
#     estimator (object *Welch*) is used with its default parameters.
# -   for each order :math:`(p,q)`, the estimation of the parameters is done by
#     maximizing the reduced equation of the Whittle likelihood function
#     (\[lik2\]), thanks to the method `build` of the object
#     :class:`~openturns.WhittleFactory`. This method applies to a time series or a process
#     sample. If the user wants to get the quantified criteria
#     :math:`AIC_c`, `AIC` and `BIC` of the model :math:`ARMA(p,q)`, he has to specify
#     it by giving a `Point` of size 0 as input parameter of
#     the method `build`.
# -   the output of the estimation is, in all the cases, one unique ARMA:
#     the ARMA with the specified order or the optimal one with respect to
#     the :math:`AIC_c` criterion.
# -   in the case of a range :math:`Ind_p \times Ind_q`, the user can get all
#     the estimated models thanks to the method *getHistory* of the object
#     class:`~openturns.WhittleFactory`. If the `build` has been parameterized by a Point
#     of size 0, the user also has access to all the quantified criteria.
#
# The synthetic data is generated using the following 1-d ARMA process:
#
# .. math::
#     X_{0,t} + 0.4 X_{0,t-1} + 0.3 X_{0,t-2} + 0.2 X_{0,t-3} + 0.1 X_{0,t-4} = E_{0,t} + 0.4 E_{0,t-1} + 0.3 E_{0,t-2}
#
# with the noise :math:`E` defined as:
#
# .. math::
#     E \sim Triangular(-1, 0, 1)
#

# %%
import openturns as ot

# %%
# Create an ARMA process
tMin = 0.0
n = 1000
timeStep = 0.1
myTimeGrid = ot.RegularGrid(tMin, timeStep, n)

myWhiteNoise = ot.WhiteNoise(ot.Triangular(-1.0, 0.0, 1.0), myTimeGrid)
myARCoef = ot.ARMACoefficients([0.4, 0.3, 0.2, 0.1])
myMACoef = ot.ARMACoefficients([0.4, 0.3])
arma = ot.ARMA(myARCoef, myMACoef, myWhiteNoise)

tseries = ot.TimeSeries(arma.getRealization())

# Create a sample of `N` time series from the process
N = 100
sample = arma.getSample(N)

# %%
# CASE 1 : we specify a (p,q) order

# Specify the order (p,q)
p = 4
q = 2

# Create the estimator
factory = ot.WhittleFactory(p, q)
print("Default spectral model factory = ", factory.getSpectralModelFactory())

# To set the spectral model factory
# For example, set :class:`~openturns.WelchFactory` as :class:`~openturns.SpectralModelFactory`
# with the Hann filtering window
# The Welch estimator splits the time series in four blocs without overlap
myFilteringWindow = ot.Hann()
mySpectralFactory = ot.WelchFactory(myFilteringWindow, 4, 0)
factory.setSpectralModelFactory(mySpectralFactory)
print("New spectral model factory = ", factory.getSpectralModelFactory())

# Estimate the ARMA model from a time series
# To get the quantified AICc, AIC and BIC criteria
arma42, criterion = factory.buildWithCriteria(tseries)
AICc, AIC, BIC = criterion[0:3]
print("AICc=", AICc, "AIC=", AIC, "BIC=", BIC)
arma42

# %%
# CASE 2 : we specify a range of (p,q) orders

# Range for p
pIndices = [1, 2, 4]
# Range for q = [4,5,6]
qIndices = [4, 5, 6]

# Build a Whittle factory with default SpectralModelFactory (WelchFactory)
# this time using ranges of order p and q
factory_range = ot.WhittleFactory(pIndices, qIndices)

# Estimate the arma model from a process sample
arma_range, criterion = factory_range.buildWithCriteria(sample)
AICc, AIC, BIC = criterion[0:3]
print("AICc=", AICc, "AIC=", AIC, "BIC=", BIC)
arma_range

# %%
# Results exploitation

# Get the white noise of the (best) estimated ARMA
arma_range.getWhiteNoise()



#! /usr/bin/env python

import openturns as ot
import openturns.testing as ott

distParams = []
distParams.append(ot.ArcsineMuSigma(8.4, 2.25))
distParams.append(ot.BetaMuSigma(0.2, 0.6, -1, 2))
distParams.append(ot.GammaMuSigma(1.5, 2.5, -0.5))
distParams.append(ot.GumbelLambdaGamma(0.6, 6.0))
distParams.append(ot.GumbelMuSigma(1.5, 1.3))
distParams.append(ot.LogNormalMuErrorFactor(0.63, 1.5, -0.5))
distParams.append(ot.LogNormalMuSigma(0.63, 3.3, -0.5))
distParams.append(ot.LogNormalMuSigmaOverMu(0.63, 5.24, -0.5))
distParams.append(ot.UniformMuSigma(8.4, 2.25))
distParams.append(ot.WeibullMaxMuSigma(1.3, 1.23, 3.1))
distParams.append(ot.WeibullMinMuSigma(1.3, 1.23, -0.5))
# defaults
distParams.append(ot.ArcsineMuSigma())
distParams.append(ot.BetaMuSigma())
distParams.append(ot.GammaMuSigma())
distParams.append(ot.GumbelLambdaGamma())
distParams.append(ot.GumbelMuSigma())
distParams.append(ot.LogNormalMuErrorFactor())
distParams.append(ot.LogNormalMuSigma())
distParams.append(ot.LogNormalMuSigmaOverMu())
distParams.append(ot.UniformMuSigma())
distParams.append(ot.WeibullMaxMuSigma())
distParams.append(ot.WeibullMinMuSigma())

for distParam in distParams:
    print("Distribution Parameters ", repr(distParam))
    print("Distribution Parameters ", distParam)

    p = distParam.getValues()
    desc = distParam.getDescription()
    print("p=", p, desc)
    p_native = distParam.evaluate()
    print("native=", p_native)
    p_roundtrip = distParam.inverse(p_native)
    print("roundtrip=", p_roundtrip)
    ott.assert_almost_equal(p_roundtrip, p)
    dist = distParam.getDistribution()
    print("built dist=", dist)
    assert dist.getParameter() == p_native

    # derivative of the native parameters with regards the parameters of the
    # distribution
    grad = distParam.gradient()
    print(grad)

    # by the finite difference technique
    eps = 1e-5
    dim = len(p)
    grad_fd = ot.SquareMatrix(ot.IdentityMatrix(dim))

    for i in range(dim):
        for j in range(dim):
            xp = list(p)
            xp[i] += eps
            xm = list(p)
            xm[i] -= eps
            grad_fd[i, j] = 0.5 * (distParam(xp)[j] - distParam(xm)[j]) / eps
    print(grad_fd)
    ott.assert_almost_equal(grad, grad_fd)



import openturns as ot
import openturns.viewer as otv

sample = ot.Gamma(6.0, 1.0).getSample(100)
ks = ot.KernelSmoothing()
bandwidth = [0.9]
fittedDist = ks.build(sample, bandwidth)

graph = fittedDist.drawPDF()
graph.add(ot.Gamma(6.0, 1.0).drawPDF())
graph.setLegends(["KS dist", "Gamma"])
otv.View(graph)



import openturns as ot
from math import exp
from matplotlib import pyplot as plt
import openturns.viewer as otv


def C(s, t):
    return exp(-4.0 * abs(s - t) / (1 + (s * s + t * t)))


N = 64
a = 4.0
# myMesh = ot.IntervalMesher([N]).build(ot.Interval(-a, a))
myMesh = ot.RegularGrid(-a, 2 * a / N, N + 1)

vertices = myMesh.getVertices()
myCovariance = ot.CovarianceMatrix(len(vertices))
for k in range(len(vertices)):
    t = vertices[k]
    for ll in range(k + 1):
        s = vertices[ll]
        myCovariance[k, ll] = C(s[0], t[0])

covarianceModel = ot.UserDefinedCovarianceModel(myMesh, myCovariance)
cov_graph = covarianceModel.draw(0, 0, -a, a, 512)
cov_graph.setTitle("User defined covariance model")

fig = plt.figure(figsize=(10, 4))
cov_axis = fig.add_subplot(111)
otv.View(cov_graph, figure=fig, axes=[cov_axis], add_legend=False, square_axes=True)



"""
Analyse the central tendency of a cantilever beam
=================================================
"""

# %%
# In this example we perform a central tendency analysis of a random variable Y using the various methods available.
# We consider the :ref:`cantilever beam <use-case-cantilever-beam>` example and show
# how to use the `TaylorExpansionMoments` and `ExpectationSimulationAlgorithm` classes.

# %%
from openturns.usecases import cantilever_beam
import openturns as ot
import openturns.viewer as otv


# %%
# We first load the data class from the usecases module :
cb = cantilever_beam.CantileverBeam()

# %%
# We want to create the random variable of interest :math:`Y=g(X)` where :math:`g(.)` is the physical model and :math:`X` is the input vectors.
#
# We create the input parameters distribution and make a random vector.
# For the sake of this example, we consider an independent copula.

distribution = ot.JointDistribution([cb.E, cb.F, cb.L, cb.II])
X = ot.RandomVector(distribution)
X.setDescription(["E", "F", "L", "I"])

# %%
# `f` is the cantilever beam model :
f = cb.model

# %%
# The random variable of interest Y is then
Y = ot.CompositeRandomVector(f, X)
Y.setDescription("Y")

# %%
# Taylor expansion
# ----------------

# %%
# Perform Taylor approximation to get the expected value of Y and the importance factors.

# %%
taylor = ot.TaylorExpansionMoments(Y)
taylor_mean_fo = taylor.getMeanFirstOrder()
taylor_mean_so = taylor.getMeanSecondOrder()
taylor_cov = taylor.getCovariance()
taylor_if = taylor.getImportanceFactors()
print("model evaluation calls number=", f.getGradientCallsNumber())
print("model gradient calls number=", f.getGradientCallsNumber())
print("model hessian calls number=", f.getHessianCallsNumber())
print("taylor mean first order=", taylor_mean_fo)
print("taylor variance=", taylor_cov)
print("taylor importance factors=", taylor_if)

# %%
graph = taylor.drawImportanceFactors()
view = otv.View(graph)

# %%
# We see that, at first order, the variable :math:`F` explains about 70% of the variance of the output :math:`Y`.
# On the other hand, the variable :math:`E` is the least significant in the variance of the output: :math:`E` only explains about 5% of the output variance.

# %%
# Monte-Carlo simulation
# ----------------------

# %%
# Perform a Monte Carlo simulation of Y to estimate its mean.

# %%
algo = ot.ExpectationSimulationAlgorithm(Y)
algo.setMaximumOuterSampling(1000)
algo.setCoefficientOfVariationCriterionType("NONE")
algo.run()
print("model evaluation calls number=", f.getEvaluationCallsNumber())
expectation_result = algo.getResult()
expectation_mean = expectation_result.getExpectationEstimate()
print(
    "monte carlo mean=",
    expectation_mean,
    "var=",
    expectation_result.getVarianceEstimate(),
)

# %%
# Central dispersion analysis based on a sample
# ---------------------------------------------

# %%
# Directly compute statistical moments based on a sample of Y. Sometimes the probabilistic model is not available and the study needs to start from the data.

# %%
Y_s = Y.getSample(1000)
y_mean = Y_s.computeMean()
y_stddev = Y_s.computeStandardDeviation()
y_quantile_95p = Y_s.computeQuantilePerComponent(0.95)
print("mean=", y_mean, "stddev=", y_stddev, "quantile@95%", y_quantile_95p)

# %%
graph = ot.KernelSmoothing().build(Y_s).drawPDF()
graph.setTitle("Kernel smoothing approximation of the output distribution")
view = otv.View(graph)

# %%
otv.View.ShowAll()



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

distribution = ot.WeibullMax(1.0, 2.5, -1.0)
size = 10000
sample = distribution.getSample(size)
factory = ot.WeibullMaxFactory()
estimatedDistribution = factory.build(sample)
print("distribution=", repr(distribution))
print("Estimated distribution=", repr(estimatedDistribution))
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedDistribution = factory.build(distribution.getParameter())
print("Distribution from parameters=", estimatedDistribution)
estimatedWeibull = factory.buildAsWeibullMax(sample)
print("Estimated weibull=", estimatedWeibull)
estimatedWeibull = factory.buildAsWeibullMax()
print("Default weibull=", estimatedWeibull)
estimatedWeibull = factory.buildAsWeibullMax(distribution.getParameter())
print("WeibullMax from parameters=", estimatedWeibull)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

# Discrete related functions
# rDiscrete
p = [0.1, 0.2, 0.3, 0.4]
print("rDiscrete(", ot.Point(p), ")=", ot.DistFunc.rDiscrete(p))
print("rDiscrete(", ot.Point(p), ", 100)=", ot.DistFunc.rDiscrete(p, 100))



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()
ot.PlatformInfo.SetNumericalPrecision(3)


distributions = [ot.Student(3.5, 2.5, 1.0)]
R = ot.CorrelationMatrix([[1.0, 0.5], [0.5, 1.0]])
distributions.append(ot.Student(4.5, [2.5] * 2, [1.5] * 2, R))
R = ot.CorrelationMatrix(10)
mu = ot.Point(10)
sigma = ot.Point(10)
for i in range(10):
    mu[i] = 0.5 * i
    sigma[i] = 0.5 + i
    for j in range(i):
        R[i, j] = 1.0 / (i + j + 2.0) ** 2.0
distributions.append(ot.Student(4.5, mu, sigma, R))
size = 10000
factory = ot.StudentFactory()
for distribution in distributions:
    sample = distribution.getSample(size)
    estimatedDistribution = factory.build(sample)
    print("distribution=", distribution)
    print("Estimated distribution=", estimatedDistribution)
estimatedDistribution = factory.build()
print("Default distribution=", estimatedDistribution)
estimatedStudent = factory.buildAsStudent(sample)
print("Student          =", distribution)
print("Estimated student=", estimatedStudent)
estimatedStudent = factory.buildAsStudent()
print("Default student=", estimatedStudent)



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()


# Dimension of test
dimension = 2

# function
function = ot.SymbolicFunction("x", "sin(x)")

# Init of the function
myFunction = ot.InverseTrendEvaluation(function)
myFunction.setName("myInverseTrendFunction")

# We check that for a Point with same values, the function is
# done by component
inPoint = ot.Point(dimension, 30.0)

# result point of the function
outPoint = myFunction(inPoint)
print("myFunction=", myFunction)
print(myFunction.getName(), "( ", inPoint, " ) = ", outPoint)

# Creation of a Sample
size = 10
inSample = ot.Sample(size, dimension)
for index in range(size):
    inSample[index] = ot.Point(dimension, index + 1)

# result of the function
outSample = myFunction(inSample)
print(myFunction.getName(), "( ", repr(inSample), " ) = ", repr(outSample))



#! /usr/bin/env python

import openturns as ot

ot.TESTPREAMBLE()

dim = 2
transformation = ot.NatafIndependentCopulaHessian(dim)
print("transformation=", repr(transformation))
point = ot.Point(dim, 0.75)
print("transformation(", point, ")=", repr(transformation.hessian(point)))
print("input dimension=", transformation.getInputDimension())
print("output dimension=", transformation.getOutputDimension())



import openturns as ot
import openturns.viewer as otv

N = 100
M = 1000
P = 10
mean = ot.SymbolicFunction("x", "sign(x)")
cov = ot.SquaredExponential([1.0], [0.1])
mesh = ot.IntervalMesher([N]).build(ot.Interval(-2.0, 2.0))
process = ot.GaussianProcess(ot.TrendTransform(mean, mesh), cov, mesh)
sample = process.getSample(M)
algo = ot.KarhunenLoeveSVDAlgorithm(sample, 1e-6)
algo.run()
result = algo.getResult()
trend = ot.TrendTransform(ot.P1LagrangeEvaluation(sample.computeMean()), mesh)
sample2 = process.getSample(P)
sample2.setName("reduction of sign(x) w/o trend")
reduced1 = ot.KarhunenLoeveReduction(result)(sample2)
reduced2 = ot.KarhunenLoeveReduction(result, trend)(sample2)
g = sample2.drawMarginal(0)
g.setColors(["red"])
g1 = reduced1.drawMarginal(0)
g1.setColors(["blue"])
drs = g1.getDrawables()
for i, d in enumerate(drs):
    d.setLineStyle("dashed")
    drs[i] = d
g1.setDrawables(drs)
g.add(g1)
g2 = reduced2.drawMarginal(0)
g2.setColors(["green"])
drs = g2.getDrawables()
for i, d in enumerate(drs):
    d.setLineStyle("dotted")
    drs[i] = d
g2.setDrawables(drs)
g.add(g2)
otv.View(g)



#! /usr/bin/env python

import openturns as ot

ot.PlatformInfo.SetNumericalPrecision(3)
distribution = ot.Beta(2.3, 2.2, -1.0, 1.0)
print(f"distribution={distribution}")
sample = distribution.getSample(1000)
factory = ot.MaximumLikelihoodFactory(ot.Beta())
infDist = factory.build(sample)
print(f"estimated distribution={infDist}")

# set (a,b) out of (r, t, a, b)
factory.setKnownParameter(
    [2, 3],
    [-1.0, 1.0],
)
infDist = factory.build(sample)
print(f"estimated distribution with bounds={infDist}")

factory = ot.MaximumLikelihoodFactory(ot.Exponential())
factory.setKnownParameter([1], [0.1])
print(factory.build())
print(factory.build([3, 0]))

# test BuildEstimator
distribution = ot.Gumbel(0.5, -0.5)
sample = distribution.getSample(1000)
for isRegular in [True, False]:
    fResult = ot.MaximumLikelihoodFactory.BuildEstimator(
        ot.GumbelFactory(), sample, isRegular
    )
    paramDist = fResult.getParameterDistribution()
    print(f"estimator(isRegular={isRegular}): {str(paramDist)[:100]}")



#!/usr/bin/env python
"""
An optimization problem with mixed discrete input variables.

Simple PLNE :

maximize     15*x + 12*y + 4*z + 2*t
subject to   -(8*x + 5*y + 3*z + 2*t - 10) >= 0

where x, y, z, t in {0,1}
Solution is [0, 1, 1, 1] with objectif value = 18

Without discrete constraint, we get :
x*= [0.625,1,0,0], f(x*)= [-21.375]

References
https://ocw.mit.edu/courses/sloan-school-of-management/15-053-optimization-methods-in-management-science-spring-2013/tutorials/MIT15_053S13_tut10.pdf
"""

import openturns as ot
import openturns.testing as ott

# Define the objective function
objectiveFun = ot.SymbolicFunction(["x", "y", "z", "t"], ["-(15*x + 12*y + 4*z + 2*t)"])
constraintFun = ot.SymbolicFunction(
    ["x", "y", "z", "t"], ["-(8*x + 5*y + 3*z + 2*t -10)"]
)
x = [0, 1, 1, 1]
print(f"f(x)={objectiveFun(x)}")
print(f"g(x)={constraintFun(x)}")

# Define problem
problem = ot.OptimizationProblem(objectiveFun)
problem.setInequalityConstraint(constraintFun)
bounds = ot.Interval([0.0] * 4, [1.0] * 4)
problem.setBounds(bounds)
problem.setMinimization(True)
problem.setVariablesType(
    [
        ot.OptimizationProblemImplementation.BINARY,
        ot.OptimizationProblemImplementation.BINARY,
        ot.OptimizationProblemImplementation.BINARY,
        ot.OptimizationProblemImplementation.BINARY,
    ]
)

# Define OptimizationAlgorithm
x0 = [0.0] * 4
algo = ot.Bonmin(problem)
algo.setStartingPoint(x0)
algo.setMaximumCallsNumber(10000)
algo.setMaximumIterationNumber(1000)


for name in ot.Bonmin.GetAlgorithmNames():
    print(f"-- {name} algorithm...")
    algo.setAlgorithmName(name)
    algo.run()

    # Retrieve result
    result = algo.getResult()
    x_star = result.getOptimalPoint()
    print("x*=", x_star)
    y_star = result.getOptimalValue()
    neval = result.getCallsNumber()
    print(f"f(x*)={y_star} neval={neval}")
    print(f"g(x*)={constraintFun(x_star)}")
    ott.assert_almost_equal(x_star, [0, 1, 1, 1], 1, 5e-4)



#! /usr/bin/env python

import openturns as ot
from math import pi

ot.TESTPREAMBLE()

# Log.Show( Log.Flags() | Log.INFO )

# Problem parameters
dimension = 3
a = 7.0
b = 0.1
# Reference analytical values
meanTh = a / 2
covTh = (b**2 * pi**8) / 18.0 + (b * pi**4) / 5.0 + (a**2) / 8.0 + 1.0 / 2.0
sob_1 = [
    (b * pi**4 / 5.0 + b**2 * pi**8 / 50.0 + 1.0 / 2.0) / covTh,
    (a**2 / 8.0) / covTh,
    0.0,
]
sob_2 = [0.0, (b**2 * pi**8 / 18.0 - b**2 * pi**8 / 50.0) / covTh, 0.0]
sob_3 = [0.0]
sob_T1 = [
    sob_1[0] + sob_2[0] + sob_2[1] + sob_3[0],
    sob_1[1] + sob_2[0] + sob_2[2] + sob_3[0],
    sob_1[2] + sob_2[1] + sob_2[2] + sob_3[0],
]
sob_T2 = [
    sob_2[0] + sob_2[1] + sob_3[0],
    sob_2[0] + sob_2[2] + sob_3[0],
    sob_2[1] + sob_2[2] + sob_3[0],
]
sob_T3 = [sob_3[0]]
# Create the Ishigami function
model = ot.SymbolicFunction(
    ["xi1", "xi2", "xi3"],
    [
        "sin(xi1) + ("
        + str(a)
        + ") * (sin(xi2)) ^ 2 + ("
        + str(b)
        + ") * xi3^4 * sin(xi1)"
    ],
)

# Create the input distribution
distribution = ot.JointDistribution([ot.Uniform(-pi, pi)] * dimension)

# Create the orthogonal basis
polynomialCollection = ot.PolynomialFamilyCollection(dimension)
for i in range(dimension):
    polynomialCollection[i] = ot.OrthogonalUniVariatePolynomialFamily(
        ot.LegendreFactory()
    )
enumerateFunction = ot.LinearEnumerateFunction(dimension)
productBasis = ot.OrthogonalBasis(
    ot.OrthogonalProductPolynomialFactory(polynomialCollection, enumerateFunction)
)

# design experiment
samplingSize = 75
experiment = ot.LowDiscrepancyExperiment(ot.SobolSequence(), distribution, samplingSize)

# generate samples
x = experiment.generate()
y = model(x)

# iso transfo
xToU = ot.DistributionTransformation(distribution, productBasis.getMeasure())

u = xToU(x)

# build basis
degree = 10
basisSize = enumerateFunction.getStrataCumulatedCardinal(degree)
basis = [productBasis.build(i) for i in range(basisSize)]

# run algorithm
factory = ot.BasisSequenceFactory(ot.LARS())
print("factory = ", factory)

seq = factory.build(u, y, basis, list(range(basisSize)))

first = 20
if seq.getSize() >= first:
    print("first ", first, " indices = ", seq.getIndices(first - 1))



import openturns as ot
from matplotlib import pyplot as plt
import openturns.viewer as otv

# Define the vertices of the mesh
vertices = [[0.0, 0.0], [1.0, 0.0], [1.0, 1.0], [1.5, 1.0], [2.0, 1.5], [0.5, 1.5]]
# Define the simplices of the mesh
simplices = [[0, 1, 2], [1, 2, 3], [2, 3, 4], [2, 4, 5], [0, 2, 5]]
# Create the Mesh
mesh2D = ot.Mesh(vertices, simplices)
# Create a Graph
graph = ot.Graph("Mesh 2D", "", "")
graph.setLegendPosition("lower right")
graph.add(mesh2D.draw())

# Then, draw it
fig = plt.figure(figsize=(4, 4))
axis = fig.add_subplot(111)
otv.View(graph, figure=fig, axes=[axis], add_legend=True)
axis.set_xlim(auto=True)



# this resets OT between each example


def reset_ot(gallery_conf, fname):
    import openturns as ot

    ot.Log.Show(ot.Log.NONE)
    ot.ResourceMap.Reset()
    ot.RandomGenerator.Reset()

    # explicitly set agg backend
    import matplotlib

    matplotlib_backend = matplotlib.get_backend().lower()
    if matplotlib_backend != "agg":
        matplotlib.use("agg")
